import numpy as np


def autocorrelation_sokal(chain: np.ndarray) -> np.ndarray:
    """
    Estimate the integrated autocorrelation time of a MCMC chain.

    Uses Sokal's adaptive truncated periodogram estimator.

    - Haario, H., Laine, M., Mira, A. et al. DRAM: Efficient
    adaptive MCMC. Stat Comput 16, 339–354 (2006).
    https://doi.org/10.1007/s11222-006-9438-0

    - Sokal A. (1997) Monte Carlo Methods in Statistical Mechanics:
    Foundations and New Algorithms. In: DeWitt-Morette C.,
    Cartier P., Folacci A. (eds) Functional Integration.
    NATO ASI Series (Series B: Physics), vol 361. Springer, Boston, MA

    Parameters
    ----------
        chain: The MCMC chain.

    Returns
    -------
        tau_est: An estimate of the integrated autocorrelation time of
        the MCMC chain.
    """
    nsamples, npar = chain.shape
    tau_est = np.zeros((npar))

    # Calculate the fast Fourier transform
    x = np.fft.fft(chain, axis=0)
    # Get the real part
    xr = np.real(x)
    # Get the imaginary part
    xi = np.imag(x)

    xr = xr**2 + xi**2
    # First value to zero
    xr[0, :] = 0.0
    # Calculate the fast Fourier transform
    # of the transformation
    xr = np.real(np.fft.fft(xr, axis=0))
    # Calculate the variance
    var = xr[0] / nsamples / (nsamples - 1)

    # Loop over parameters
    for j in range(npar):
        if var[j] == 0.0:
            continue
        # Normalize by the first value
        xr[:, j] = xr[:, j] / xr[0, j]
        # Initiate variable
        _sum = -1 / 3
        # Loop over samples
        for i in range(nsamples):
            _sum = _sum + xr[i, j] - 1 / 6
            if _sum < 0.0:
                tau_est[j] = 2 * (_sum + i / 6)
                break

    return tau_est



"""Utility functions for :py:func:`pypesto.optimize.minimize`."""

import logging
import os
from pathlib import Path
from typing import Iterable, List

import h5py

from .. import C
from ..engine import Engine, SingleCoreEngine
from ..history import CsvHistoryTemplateError, HistoryOptions, HistoryTypeError
from ..result import Result
from .optimizer import OptimizerResult

logger = logging.getLogger(__name__)


def preprocess_hdf5_history(
    history_options: HistoryOptions,
    engine: Engine,
):
    """Create a folder for partial HDF5 files if parallelization is used.

    This is because single hdf5 file access is not thread-safe.

    Parameters
    ----------
    engine:
        The Engine which is used in the optimization.
    history_options:
        The HistoryOptions used in the optimization.

    Returns
    -------
    history_requires_postprocessing:
        Whether history storage post-processing is required.
    """
    storage_file = history_options.storage_file

    # nothing to do if no history stored
    if storage_file is None:
        return False

    # extract storage type
    path = Path(storage_file)

    # nothing to do if csv history and correctly set
    if path.suffix[1:] in C.SUFFIXES_CSV:
        if "{id}" not in storage_file:
            raise CsvHistoryTemplateError(storage_file)
        return False

    # assuming hdf5 history henceforth
    if path.suffix[1:] not in C.SUFFIXES_HDF5:
        raise HistoryTypeError(path.suffix)

    # nothing to do if no parallelization
    if isinstance(engine, SingleCoreEngine):
        return False

    # create directory with same name as original file stem
    template_path = (
        path.parent / path.stem / (path.stem + "_{id}" + path.suffix)
    )
    template_path.parent.mkdir(parents=True, exist_ok=True)
    # set history file to template path
    history_options.storage_file = str(template_path)

    return True


def postprocess_hdf5_history(
    ret: List[OptimizerResult],
    storage_file: str,
    history_options: HistoryOptions,
) -> None:
    """Create single history file pointing to files of multiple starts.

    Create links in `storage_file` to the history of each start contained in
    `ret`, the results of the optimization.

    Parameters
    ----------
    ret:
        The result iterable returned by the optimization.
    storage_file:
        The filename of the hdf5 file in which the histories
        are to be gathered.
    history_options:
        History options used in the optimization.
    """
    # create hdf5 file that gathers the others within history group
    with h5py.File(storage_file, mode='w') as f:
        # create file and group
        f.require_group("history")
        # append links to each single result file
        for result in ret:
            id = result['id']
            f[f'history/{id}'] = h5py.ExternalLink(
                result['history'].file, f'history/{id}'
            )

    # reset storage file (undo preprocessing changes)
    history_options.storage_file = storage_file


def bound_n_starts_from_env(n_starts: int):
    """Bound number of optimization starts from environment variable.

    Uses environment variable `PYPESTO_MAX_N_STARTS`.
    This is used to speed up testing, while in application it should not
    be used.

    Parameters
    ----------
    n_starts: Number of starts desired.

    Returns
    -------
    n_starts_new:
        The original number of starts, or the minimum with the environment
        variable, if exists.
    """
    if C.PYPESTO_MAX_N_STARTS not in os.environ:
        return n_starts
    n_starts_new = min(n_starts, int(os.environ[C.PYPESTO_MAX_N_STARTS]))

    logger.info(
        f"Bounding number of samples from {n_starts} to {n_starts_new} via "
        f"environment variable {C.PYPESTO_MAX_N_STARTS}"
    )

    return n_starts_new


def assign_ids(
    n_starts: int,
    ids: Iterable[str] = None,
    result: Result = None,
) -> Iterable[str]:
    """
    Assign ids to starts.

    ids:
        Ids assigned to the startpoints.
    result:
        A result object to append the optimization results to. For example,
        one might append more runs to a previous optimization. Assign only
        unique ids.
    n_starts:
        Number of starts of the optimizer.
    """
    used_ids = set()
    if result is not None:
        used_ids = set(result.optimize_result.id)
        N_used = len(used_ids)
    if ids is None:
        i = 0
        ids = [str(j) for j in range(i * n_starts, (i + 1) * n_starts)]
        while not used_ids.isdisjoint(ids):
            i += 1
            ids = [
                str(j) for j in range(N_used * i, N_used * (i + 1) + n_starts)
            ]
    if len(ids) != n_starts:
        raise AssertionError("Number of starts and ids must coincide.")
    if not used_ids.isdisjoint(ids):
        raise AssertionError(
            "Manually assigned ids must differ from existing ones."
        )
    return ids



from __future__ import annotations

import abc
import logging
import re
import time
from typing import TYPE_CHECKING, Dict, Optional

import numpy as np
import scipy.optimize

from ..C import FVAL, GRAD, MODE_FUN, MODE_RES
from ..history import (
    HistoryOptions,
    NoHistory,
    OptimizerHistory,
    create_history,
)
from ..objective import Objective
from ..problem import Problem
from ..result import OptimizerResult
from .load import fill_result_from_history
from .options import OptimizeOptions

if TYPE_CHECKING:
    import fides

logger = logging.getLogger(__name__)


class OptimizerImportError(ImportError):
    """Exception raised when an optimizer is not available."""

    def __init__(self, optimizer: str):
        super().__init__(
            f'Optimizer "{optimizer}" not available, install corresponding '
            f'package e.g. via "pip install pypesto[{optimizer}]"'
        )


def history_decorator(minimize):
    """Initialize and extract information stored in the history.

    Default decorator for the minimize() method.
    """

    def wrapped_minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ):
        if history_options is None:
            history_options = HistoryOptions()

        objective = problem.objective

        # initialize the objective
        objective.initialize()

        # initialize the history
        history = create_history(
            id=id,
            x_names=[problem.x_names[ix] for ix in problem.x_free_indices],
            options=history_options,
        )
        optimizer_history = OptimizerHistory(
            history=history,
            x0=x0,
            lb=problem.lb,
            ub=problem.ub,
        )

        # plug in history for the objective to record it
        objective.history = optimizer_history

        # perform the actual minimization
        try:
            result = minimize(
                self,
                problem=problem,
                x0=x0,
                id=id,
                history_options=history_options,
                optimize_options=optimize_options,
            )
            result.id = id
            objective.history.finalize(
                message=result.message, exitflag=result.exitflag
            )
        except Exception as err:
            if optimize_options.allow_failed_starts:
                logger.error(f'start {id} failed: {err}')
                result = OptimizerResult(
                    x0=x0, exitflag=-1, message=str(err), id=id
                )
            else:
                raise

        # maybe override results from history depending on options
        result = fill_result_from_history(
            result=result,
            optimizer_history=objective.history,
            optimize_options=optimize_options,
        )

        # clean up, history is available from result
        objective.history = NoHistory()

        return result

    return wrapped_minimize


def time_decorator(minimize):
    """Measure time of optimization.

    Default decorator for the minimize() method to take time.
    Currently, the method time.time() is used, which measures
    the wall-clock time.
    """

    def wrapped_minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ):
        start_time = time.time()
        result = minimize(
            self,
            problem=problem,
            x0=x0,
            id=id,
            history_options=history_options,
            optimize_options=optimize_options,
        )
        used_time = time.time() - start_time
        result.time = used_time
        return result

    return wrapped_minimize


def fix_decorator(minimize):
    """Include also fixed parameters in the result arrays of minimize().

    Default decorator for the minimize() method (nans will be inserted in the
    derivatives).
    """

    def wrapped_minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ):
        # perform the actual optimization
        result = minimize(
            self,
            problem=problem,
            x0=x0,
            id=id,
            history_options=history_options,
            optimize_options=optimize_options,
        )

        # vectors to full vectors
        result.update_to_full(problem)

        logger.debug(
            f"Final fval={result.fval:.4f}, time={result.time:.4f}s, "
            f"n_fval={result.n_fval}.",
        )

        return result

    return wrapped_minimize


class Optimizer(abc.ABC):
    """
    Optimizer base class, not functional on its own.

    An optimizer takes a problem, and possibly a start point, and then
    performs an optimization. It returns an OptimizerResult.
    """

    def __init__(self):
        """Initialize base class."""

    @abc.abstractmethod
    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """
        Perform optimization.

        Parameters
        ----------
        problem:
            The problem to find optimal parameters for.
        x0:
            The starting parameters.
        id:
            Multistart id.
        history_options:
            Optimizer history options.
        optimize_options:
            Global optimization options.
        """

    @abc.abstractmethod
    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False

    def get_default_options(self):
        """Create default options specific for the optimizer."""
        return None


def check_finite_bounds(lb, ub):
    """Raise if bounds are not finite."""
    if not np.isfinite(lb).all() or not np.isfinite(ub).all():
        raise ValueError(
            'Selected optimizer cannot work with unconstrained '
            'optimization problems.'
        )


class ScipyOptimizer(Optimizer):
    """
    Use the SciPy optimizers.

    Find details on the optimizer and configuration options at:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.\
        optimize.minimize.html#scipy.optimize.minimize
    """

    def __init__(
        self,
        method: str = 'L-BFGS-B',
        tol: float = None,
        options: Dict = None,
    ):
        super().__init__()

        self.method = method

        self.options = options
        if self.options is None:
            self.options = ScipyOptimizer.get_default_options(self)
        self.tol = tol

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__} method={self.method}"
        # print everything that is customized
        if self.tol is not None:
            rep += f" tol={self.tol}"
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        lb = problem.lb
        ub = problem.ub
        objective = problem.objective

        if self.is_least_squares():
            # set tolerance to default of scipy optimizer
            tol = self.tol
            if tol is None:
                tol = 1e-8
            # is a residual based least squares method
            if not objective.has_res:
                raise Exception(
                    "For least squares optimization, the objective "
                    "must be able to compute residuals."
                )

            ls_method = self.method[3:]
            bounds = (lb, ub)

            fun = objective.get_res
            jac = objective.get_sres if objective.has_sres else '2-point'
            # TODO: pass jac computing methods in options

            if self.options is not None:
                ls_options = self.options.copy()
                ls_options['verbose'] = (
                    2
                    if 'disp' in ls_options.keys() and ls_options['disp']
                    else 0
                )
                ls_options.pop('disp', None)
                ls_options['max_nfev'] = ls_options.pop('maxfun', None)
            else:
                ls_options = {}

            # optimize
            res = scipy.optimize.least_squares(
                fun=fun,
                x0=x0,
                method=ls_method,
                jac=jac,
                bounds=bounds,
                tr_solver=ls_options.pop(
                    'tr_solver', 'lsmr' if len(x0) > 1 else 'exact'
                ),
                loss='linear',
                ftol=tol,
                **ls_options,
            )
            # extract fval/grad from result, note that fval is not available
            # from least squares solvers
            grad = getattr(res, 'grad', None)
            fval = None
        else:
            # is an fval based optimization method

            if not objective.has_fun:
                raise Exception(
                    "For this optimizer, the objective must "
                    "be able to compute function values"
                )

            bounds = scipy.optimize.Bounds(lb, ub)

            # fun_may_return_tuple = self.method.lower() in \
            #    ['cg', 'bfgs', 'newton-cg', 'l-bfgs-b', 'tnc', 'slsqp',
            #        'dogleg', 'trust-ncg']
            # TODO: is it more efficient to have tuple as output of fun?
            method_supports_grad = self.method.lower() in [
                'cg',
                'bfgs',
                'newton-cg',
                'l-bfgs-b',
                'tnc',
                'slsqp',
                'dogleg',
                'trust-ncg',
                'trust-krylov',
                'trust-exact',
                'trust-constr',
            ]
            method_supports_hess = self.method.lower() in [
                'newton-cg',
                'dogleg',
                'trust-ncg',
                'trust-krylov',
                'trust-exact',
                'trust-constr',
            ]
            method_supports_hessp = self.method.lower() in [
                'newton-cg',
                'trust-ncg',
                'trust-krylov',
                'trust-constr',
            ]
            # switch off passing over functions if not applicable (e.g.
            # NegLogParameterPrior) since grad/hess attributes do not exist
            if not isinstance(objective, Objective):
                if not hasattr(objective, 'grad'):
                    objective.grad = False
                if not hasattr(objective, 'hess'):
                    objective.hess = False
            # Todo Resolve warning by implementing saving of hess temporarily
            #  in objective and pass to scipy seperately
            if objective.hess is True:
                logger.warning(
                    "scipy.optimize.minimize does not support "
                    "passing fun and hess as one function. Hence "
                    "for each evaluation of hess, fun will be "
                    "evaluated again. This can lead to increased "
                    "computation times. If possible, separate fun "
                    "and hess."
                )
            if objective.grad is True:

                def fun(x):
                    return objective(x, sensi_orders=(0, 1))

            else:
                fun = objective.get_fval
            jac = (
                objective.get_grad
                if objective.has_grad
                and method_supports_grad
                and objective.grad is not True
                else (True if objective.grad is True else None)
            )
            hess = (
                objective.get_hess
                if objective.has_hess and method_supports_hess
                else None
            )
            hessp = (
                objective.get_hessp
                if objective.has_hessp and method_supports_hessp
                else None
            )
            # minimize will ignore hessp otherwise
            if hessp is not None:
                hess = None

            # optimize
            res = scipy.optimize.minimize(
                fun=fun,
                x0=x0,
                method=self.method,
                jac=jac,
                hess=hess,
                hessp=hessp,
                bounds=bounds,
                options=self.options,
                tol=self.tol,
            )
            # extract fval/grad from result
            grad = getattr(res, 'jac', None)
            fval = res.fun

        # fill in everything known, although some parts will be overwritten
        optimizer_result = OptimizerResult(
            x=np.array(res.x),
            fval=fval,
            grad=grad,
            hess=getattr(res, 'hess', None),
            exitflag=res.status,
            message=res.message,
        )

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return re.match(r'(?i)^(ls_)', self.method)

    def get_default_options(self):
        """Create default options specific for the optimizer."""
        if self.is_least_squares():
            options = {'max_nfev': 1000, 'disp': False}
        else:
            options = {'maxfun': 1000, 'disp': False}
        return options


class IpoptOptimizer(Optimizer):
    """Use IpOpt (https://pypi.org/project/ipopt/) for optimization."""

    def __init__(self, options: Dict = None):
        """
        Initialize.

        Parameters
        ----------
        options:
            Options are directly passed on to `cyipopt.minimize_ipopt`.
        """
        super().__init__()
        self.options = options

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__}"
        # print everything that is customized
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        try:
            import cyipopt
        except ImportError:
            raise OptimizerImportError("ipopt")

        objective = problem.objective

        bounds = np.array([problem.lb, problem.ub]).T

        ret = cyipopt.minimize_ipopt(
            fun=objective.get_fval,
            x0=x0,
            method=None,  # ipopt does not use this argument for anything
            jac=objective.get_grad,
            hess=None,  # ipopt does not support Hessian yet
            hessp=None,  # ipopt does not support Hessian vector product yet
            bounds=bounds,
            tol=None,  # can be set via options
            options=self.options,
        )

        # the ipopt return object is a scipy.optimize.OptimizeResult
        return OptimizerResult(
            x=ret.x, exitflag=ret.status, message=ret.message
        )

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False


class DlibOptimizer(Optimizer):
    """Use the Dlib toolbox for optimization."""

    def __init__(self, options: Dict = None):
        super().__init__()

        self.options = options
        if self.options is None:
            self.options = DlibOptimizer.get_default_options(self)
        elif 'maxiter' not in self.options:
            raise KeyError('Dlib options are missing the key word ' 'maxiter.')

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__}"
        # print everything that is customized
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        lb = problem.lb
        ub = problem.ub
        check_finite_bounds(lb, ub)
        objective = problem.objective

        try:
            import dlib
        except ImportError:
            raise OptimizerImportError("dlib")

        if not objective.has_fun:
            raise ValueError(
                "For this optimizer, the objective must "
                "be able to return function values."
            )

        # dlib requires variable length arguments
        def get_fval_vararg(*x):
            return objective.get_fval(x)

        dlib.find_min_global(
            get_fval_vararg,
            list(lb),
            list(ub),
            int(self.options['maxiter']),
            0.002,
        )

        optimizer_result = OptimizerResult()

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False

    def get_default_options(self):
        """Create default options specific for the optimizer."""
        return {'maxiter': 10000}


class PyswarmOptimizer(Optimizer):
    """Global optimization using pyswarm."""

    def __init__(self, options: Dict = None):
        super().__init__()

        if options is None:
            options = {'maxiter': 200}
        self.options = options

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__}"
        # print everything that is customized
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        lb = problem.lb
        ub = problem.ub

        try:
            import pyswarm
        except ImportError:
            raise OptimizerImportError("pyswarm")

        check_finite_bounds(lb, ub)

        xopt, fopt = pyswarm.pso(
            problem.objective.get_fval, lb, ub, **self.options
        )

        optimizer_result = OptimizerResult(x=np.array(xopt), fval=fopt)

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False


class CmaesOptimizer(Optimizer):
    """
    Global optimization using cma-es.

    Package homepage: https://pypi.org/project/cma-es/
    """

    def __init__(self, par_sigma0: float = 0.25, options: Dict = None):
        """
        Initialize.

        Parameters
        ----------
        par_sigma0:
            scalar, initial standard deviation in each coordinate.
            par_sigma0 should be about 1/4th of the search domain width
            (where the optimum is to be expected)
        options:
            Optimizer options that are directly passed on to cma.
        """
        super().__init__()

        if options is None:
            options = {'maxiter': 10000}
        self.options = options
        self.par_sigma0 = par_sigma0

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__} par_sigma0={self.par_sigma0}"
        # print everything that is customized
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        lb = problem.lb
        ub = problem.ub

        check_finite_bounds(lb, ub)

        sigma0 = self.par_sigma0 * np.median(ub - lb)
        self.options['bounds'] = [lb, ub]

        try:
            import cma
        except ImportError:
            raise OptimizerImportError("cma")

        result = (
            cma.CMAEvolutionStrategy(
                x0,
                sigma0,
                inopts=self.options,
            )
            .optimize(problem.objective.get_fval)
            .result
        )

        optimizer_result = OptimizerResult(
            x=np.array(result[0]), fval=result[1]
        )

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False


class ScipyDifferentialEvolutionOptimizer(Optimizer):
    """
    Global optimization using scipy's differential evolution optimizer.

    Package homepage: https://docs.scipy.org/doc/scipy/reference/generated\
        /scipy.optimize.differential_evolution.html

    Parameters
    ----------
    options:
        Optimizer options that are directly passed on to scipy's optimizer.


    Examples
    --------
    Arguments that can be passed to options:

    maxiter:
        used to calculate the maximal number of funcion evaluations by
        maxfevals = (maxiter + 1) * popsize * len(x)
        Default: 100
    popsize:
        population size, default value 15
    """

    def __init__(self, options: Dict = None):
        super().__init__()

        if options is None:
            options = {'maxiter': 100}
        self.options = options

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__}"
        # print everything that is customized
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        bounds = list(zip(problem.lb, problem.ub))

        result = scipy.optimize.differential_evolution(
            problem.objective.get_fval, bounds, **self.options
        )

        optimizer_result = OptimizerResult(
            x=np.array(result.x), fval=result.fun
        )

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False


class PyswarmsOptimizer(Optimizer):
    """
    Global optimization using pyswarms.

    Package homepage: https://pyswarms.readthedocs.io/en/latest/index.html

    Parameters
    ----------
    par_popsize:
        number of particles in the swarm, default value 10

    options:
        Optimizer options that are directly passed on to pyswarms.
        c1: cognitive parameter
        c2: social parameter
        w: inertia parameter
        Default values are (c1,c2,w) = (0.5, 0.3, 0.9)

    Examples
    --------
    Arguments that can be passed to options:

    maxiter:
        used to calculate the maximal number of funcion evaluations.
        Default: 1000
    """

    def __init__(self, par_popsize: float = 10, options: Dict = None):
        super().__init__()

        all_options = {'maxiter': 1000, 'c1': 0.5, 'c2': 0.3, 'w': 0.9}
        if options is None:
            options = {}
        all_options.update(options)
        self.options = all_options
        self.par_popsize = par_popsize

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__} par_popsize={self.par_popsize}"
        # print everything that is customized
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        lb = problem.lb
        ub = problem.ub

        try:
            import pyswarms
        except ImportError:
            raise OptimizerImportError("pyswarms")

        # check for finite values for the bounds
        if np.isfinite(lb).all() is np.False_:
            raise ValueError(
                "This optimizer can only handle finite lower bounds."
            )
        if np.isfinite(ub).all() is np.False_:
            raise ValueError(
                "This optimizer can only handle finite upper bounds."
            )

        optimizer = pyswarms.single.global_best.GlobalBestPSO(
            n_particles=self.par_popsize,
            dimensions=len(x0),
            options=self.options,
            bounds=(lb, ub),
        )

        def successively_working_fval(swarm: np.ndarray) -> np.ndarray:
            """Evaluate the function for all parameters in the swarm object.

            Parameters
            ----------
            swarm: np.ndarray, shape (n_particales_in_swarm, n_parameters)

            Returns
            -------
            result: np.ndarray, shape (n_particles_in_swarm)
            """
            n_particles = swarm.shape[0]
            result = np.zeros(n_particles)
            # iterate over the particles in the swarm
            for i_particle, par in enumerate(swarm):
                result[i_particle] = problem.objective.get_fval(par)

            return result

        cost, pos = optimizer.optimize(
            successively_working_fval,
            iters=self.options['maxiter'],
            verbose=False,
        )

        optimizer_result = OptimizerResult(
            x=pos,
            fval=float(cost),
        )

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False


class NLoptOptimizer(Optimizer):
    """
    Global/Local optimization using NLopt.

    Package homepage: https://nlopt.readthedocs.io/en/latest/
    """

    def __init__(
        self,
        method=None,
        local_method=None,
        options: Dict = None,
        local_options: Dict = None,
    ):
        """
        Initialize.

        Parameters
        ----------
        method:
            Local or global Optimizer to use for minimization.
        local_method:
            Local method to use in combination with the global optimizer (
            for the MLSL family of solvers) or to solve a subproblem (for the
            AUGLAG family of solvers)
        options:
            Optimizer options. scipy option `maxiter` is automatically
            transformed into `maxeval` and takes precedence.
        local_options:
            Optimizer options for the local method
        """
        super().__init__()

        if options is None:
            options = {}
        elif 'maxiter' in options:
            options['maxeval'] = options.pop('maxiter')
        if local_options is None:
            local_options = {}
        self.options = options
        self.local_options = local_options

        try:
            import nlopt
        except ImportError:
            raise OptimizerImportError("nlopt")

        if method is None:
            method = nlopt.LD_LBFGS

        needs_local_method = [
            nlopt.G_MLSL,
            nlopt.G_MLSL_LDS,
            nlopt.GD_MLSL,
            nlopt.GD_MLSL_LDS,
            nlopt.AUGLAG,
            nlopt.AUGLAG_EQ,
        ]

        if local_method is None and method in needs_local_method:
            local_method = nlopt.LD_LBFGS

        if local_method is not None and method not in needs_local_method:
            raise ValueError(
                f'Method "{method}" does not allow a local '
                f'method. Please set `local_method` to None.'
            )

        self.local_methods = [
            nlopt.LD_VAR1,
            nlopt.LD_VAR2,
            nlopt.LD_TNEWTON_PRECOND_RESTART,
            nlopt.LD_TNEWTON_PRECOND,
            nlopt.LD_TNEWTON_RESTART,
            nlopt.LD_TNEWTON,
            nlopt.LD_LBFGS,
            nlopt.LD_SLSQP,
            nlopt.LD_CCSAQ,
            nlopt.LD_MMA,
            nlopt.LN_SBPLX,
            nlopt.LN_NELDERMEAD,
            nlopt.LN_PRAXIS,
            nlopt.LN_NEWUOA,
            nlopt.LN_NEWUOA_BOUND,
            nlopt.LN_BOBYQA,
            nlopt.LN_COBYLA,
        ]
        self.global_methods = [
            nlopt.GN_ESCH,
            nlopt.GN_ISRES,
            nlopt.GN_AGS,
            nlopt.GD_STOGO,
            nlopt.GD_STOGO_RAND,
            nlopt.G_MLSL,
            nlopt.G_MLSL_LDS,
            nlopt.GD_MLSL,
            nlopt.GD_MLSL_LDS,
            nlopt.GN_CRS2_LM,
            nlopt.GN_ORIG_DIRECT,
            nlopt.GN_ORIG_DIRECT_L,
            nlopt.GN_DIRECT,
            nlopt.GN_DIRECT_L,
            nlopt.GN_DIRECT_L_NOSCAL,
            nlopt.GN_DIRECT_L_RAND,
            nlopt.GN_DIRECT_L_RAND_NOSCAL,
        ]
        self.hybrid_methods = [nlopt.AUGLAG, nlopt.AUGLAG_EQ]
        methods = (
            self.local_methods + self.global_methods + self.hybrid_methods
        )

        if method not in methods:
            raise ValueError(
                f'"{method}" is not a valid method. Valid '
                f'methods are: {methods}'
            )

        self.method = method

        if local_method is not None and local_method not in self.local_methods:
            raise ValueError(
                f'"{local_method}" is not a valid method. Valid '
                f'methods are: {self.local_methods}'
            )

        self.local_method = local_method

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__} method={self.method}"
        # print everything that is customized
        if self.local_method is not None:
            rep += f" local_method={self.local_method}"
        if self.options is not None:
            rep += f" options={self.options}"
        if self.local_options is not None:
            rep += f" local_options={self.local_methods}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        import nlopt

        opt = nlopt.opt(self.method, problem.dim)

        valid_options = [
            'ftol_abs',
            'ftol_rel',
            'xtol_abs',
            'xtol_rel',
            'stopval',
            'x_weights',
            'maxeval',
            'maxtime',
            'initial_step',
        ]

        def set_options(o, options):
            for option, value in options.items():
                if option not in valid_options:
                    raise ValueError(
                        f'"{option}" is not a valid option. Valid '
                        f'options are: {valid_options}'
                    )
                getattr(o, f'set_{option}')(value)

        if self.local_method is not None:
            local_opt = nlopt.opt(self.local_method, problem.dim)
            set_options(local_opt, self.local_options)
            opt.set_local_optimizer(local_opt)

        if self.method in self.global_methods:
            check_finite_bounds(problem.ub, problem.lb)

        opt.set_lower_bounds(problem.lb)
        opt.set_upper_bounds(problem.ub)

        def nlopt_objective(x, grad):
            if grad.size > 0:
                sensi_orders = (0, 1)
            else:
                sensi_orders = (0,)
            r = problem.objective(x, sensi_orders, MODE_FUN, True)
            if grad.size > 0:
                grad[:] = r[GRAD]  # note that this must be inplace
            return r[FVAL]

        opt.set_min_objective(nlopt_objective)

        set_options(opt, self.options)
        try:
            result = opt.optimize(x0)
            msg = 'Finished Successfully.'
        except (
            nlopt.RoundoffLimited,
            nlopt.ForcedStop,
            ValueError,
            RuntimeError,
            MemoryError,
        ) as e:
            result = None
            msg = str(e)

        optimizer_result = OptimizerResult(
            x=result,
            fval=opt.last_optimum_value(),
            message=msg,
            exitflag=opt.last_optimize_result(),
        )

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False


class FidesOptimizer(Optimizer):
    """
    Global/Local optimization using the trust region optimizer fides.

    Package Homepage: https://fides-optimizer.readthedocs.io/en/latest
    """

    def __init__(
        self,
        hessian_update: Optional[
            fides.hessian_approximation.HessianApproximation
        ] = 'default',
        options: Optional[Dict] = None,
        verbose: Optional[int] = logging.INFO,
    ):
        """
        Initialize.

        Parameters
        ----------
        options:
            Optimizer options.
        hessian_update:
            Hessian update strategy. If this is None, a hybrid approximation
            that switches from the problem.objective provided Hessian (
            approximation) to a BFGS approximation will be used.
        """
        super().__init__()

        try:
            import fides
        except ImportError:
            raise OptimizerImportError("fides")

        if (
            (hessian_update is not None)
            and (hessian_update != 'default')
            and not isinstance(
                hessian_update,
                fides.hessian_approximation.HessianApproximation,
            )
        ):
            raise ValueError(
                'Incompatible type for hessian update. '
                'Must be a HessianApproximation, '
                f'was {type(hessian_update)}.'
            )

        if options is None:
            options = {}

        self.verbose = verbose
        self.options = options
        self.hessian_update = hessian_update

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__} "
        # print everything that is customized
        if self.hessian_update is not None:
            if self.hessian_update == 'default':
                rep += f" hessian_update={self.hessian_update}"
            else:
                rep += (
                    f" hessian_update="
                    f"{self.hessian_update.__class__.__name__}"
                )
        if self.verbose is not None:
            rep += f" verbose={self.verbose}"
        if self.options is not None:
            rep += f" options={self.options}"
        return rep + ">"

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
        optimize_options: OptimizeOptions = None,
    ) -> OptimizerResult:
        """Perform optimization. Parameters: see `Optimizer` documentation."""
        import fides

        if self.hessian_update == 'default':
            if not problem.objective.has_hess:
                logging.warning(
                    'Fides is using BFGS as hessian approximation, '
                    'as the problem does not provide a Hessian. '
                    'Specify a Hessian to use a more efficient '
                    'hybrid approximation scheme.'
                )
                _hessian_update = fides.BFGS()
            else:
                _hessian_update = fides.HybridFixed()
        else:
            _hessian_update = self.hessian_update

        resfun = (
            _hessian_update.requires_resfun
            if _hessian_update is not None
            else False
        )

        args = {'mode': MODE_RES if resfun else MODE_FUN}

        if not problem.objective.has_grad:
            raise ValueError(
                'Fides cannot be applied to problems '
                'with objectives that do not support '
                'gradient evaluation.'
            )

        if _hessian_update is None or (
            _hessian_update.requires_hess and not resfun
        ):
            if not problem.objective.has_hess:
                raise ValueError(
                    'Specified hessian update scheme cannot be '
                    'used with objectives that do not support '
                    'Hessian computation.'
                )
            args['sensi_orders'] = (0, 1, 2)
        else:
            args['sensi_orders'] = (0, 1)

        opt = fides.Optimizer(
            fun=problem.objective,
            funargs=args,
            ub=problem.ub,
            lb=problem.lb,
            verbose=self.verbose,
            hessian_update=_hessian_update,
            options=self.options,
            resfun=resfun,
        )

        try:
            opt.minimize(x0)
            msg = self._convert_exitflag_to_message(opt)
        except RuntimeError as err:
            msg = str(err)

        optimizer_result = OptimizerResult(
            x=opt.x_min,
            fval=opt.fval_min if not resfun else None,
            grad=opt.grad_min,
            hess=opt.hess,
            message=msg,
            exitflag=opt.exitflag,
        )

        return optimizer_result

    def is_least_squares(self):
        """Check whether optimizer is a least squares optimizer."""
        return False

    def _convert_exitflag_to_message(self, opt: fides.Optimizer):
        """
        Convert the exitflag of a run to an informative message.

        Parameters
        ----------
        opt:
            The fides.Optimizer that has finished minimizing storing the
            exitflag.

        Returns
        -------
            An informative message on the cause of termination. Based on
            fides documentation.
        """
        import fides

        messages = {
            fides.ExitFlag.DID_NOT_RUN: "Optimizer did not run",
            fides.ExitFlag.MAXITER: "Reached maximum number of allowed iterations",
            fides.ExitFlag.MAXTIME: "Expected to reach maximum allowed time in next iteration",
            fides.ExitFlag.NOT_FINITE: "Encountered non-finite fval/grad/hess",
            fides.ExitFlag.EXCEEDED_BOUNDARY: "Exceeded specified boundaries",
            fides.ExitFlag.DELTA_TOO_SMALL: "Trust Region Radius too small to proceed",
            fides.ExitFlag.FTOL: "Converged according to fval difference",
            fides.ExitFlag.XTOL: "Converged according to x difference",
            fides.ExitFlag.GTOL: "Converged according to gradient norm",
        }
        return messages.get(
            opt.exitflag, f"exitflag={opt.exitflag} is not defined in fides."
        )



"""A set of helper functions."""
import logging
import os
from typing import Tuple

import numpy as np

from ..C import PYPESTO_MAX_N_SAMPLES
from ..result import Result
from .diagnostics import geweke_test

logger = logging.getLogger(__name__)


def calculate_ci_mcmc_sample(
    result: Result,
    ci_level: float = 0.95,
    exclude_burn_in: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate parameter credibility intervals based on MCMC samples.

    Parameters
    ----------
    result:
        The pyPESTO result object with filled sample result.
    ci_level:
        Lower tail probability, defaults to 95% interval.

    Returns
    -------
    lb, ub:
        Bounds of the MCMC percentile-based confidence interval.
    """
    burn_in = 0
    if exclude_burn_in:
        # Check if burn in index is available
        if result.sample_result.burn_in is None:
            geweke_test(result)

        # Get burn in index
        burn_in = result.sample_result.burn_in

    # Get converged parameter samples as numpy arrays
    chain = np.asarray(result.sample_result.trace_x[0, burn_in:, :])

    lb, ub = calculate_ci(chain, ci_level=ci_level, axis=0)
    return lb, ub


def calculate_ci_mcmc_sample_prediction(
    simulated_values: np.ndarray,
    ci_level: float = 0.95,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate prediction credibility intervals based on MCMC samples.

    Parameters
    ----------
    simulated_values:
        Simulated model states or model observables.
    ci_level:
        Lower tail probability, defaults to 95% interval.

    Returns
    -------
    lb, ub:
        Bounds of the MCMC-based prediction confidence interval.
    """
    lb, ub = calculate_ci(simulated_values, ci_level=ci_level, axis=1)
    return lb, ub


def calculate_ci(
    values: np.ndarray,
    ci_level: float,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate confidence/credibility levels using percentiles.

    Parameters
    ----------
    values:
        The values used to calculate percentiles.
    ci_level:
        Lower tail probability.
    kwargs:
        Additional keyword arguments are passed to the `numpy.percentile` call.

    Returns
    -------
    lb, ub:
        Bounds of the confidence/credibility interval.
    """
    # Percentile values corresponding to the CI level
    percentiles = 100 * np.array([(1 - ci_level) / 2, 1 - (1 - ci_level) / 2])
    # Upper and lower bounds
    lb, ub = np.percentile(values, percentiles, **kwargs)
    return lb, ub


def bound_n_samples_from_env(n_samples: int):
    """Bound number of samples from environment variable.

    Uses environment variable `PYPESTO_MAX_N_SAMPLES`.
    This is used to speed up testing, while in application it should not
    be used.

    Parameters
    ----------
    n_samples: Number of samples desired.

    Returns
    -------
    n_samples_new:
        The original number of samples, or the minimum with the environment
        variable, if exists.
    """
    if PYPESTO_MAX_N_SAMPLES not in os.environ:
        return n_samples
    n_samples_new = min(n_samples, int(os.environ[PYPESTO_MAX_N_SAMPLES]))

    logger.info(
        f"Bounding number of samples from {n_samples} to {n_samples_new} via "
        f"environment variable {PYPESTO_MAX_N_SAMPLES}"
    )

    return n_samples_new



from typing import Dict, Union


class OptimizeOptions(dict):
    """
    Options for the multistart optimization.

    Parameters
    ----------
    allow_failed_starts:
        Flag indicating whether we tolerate that exceptions are thrown during
        the minimization process.
    report_sres:
        Flag indicating whether sres will be stored in the results object.
        Deactivating this option will improve memory consumption for large
        scale problems.
    report_hess:
        Flag indicating whether hess will be stored in the results object.
        Deactivating this option will improve memory consumption for large
        scale problems.
    history_beats_optimizer:
        Whether the optimal value recorded by pyPESTO in the history has
        priority over the optimal value reported by the optimizer (True)
        or not (False).
    """

    def __init__(
        self,
        allow_failed_starts: bool = True,
        report_sres: bool = True,
        report_hess: bool = True,
        history_beats_optimizer: bool = True,
    ):
        super().__init__()

        self.allow_failed_starts: bool = allow_failed_starts
        self.report_sres: bool = report_sres
        self.report_hess: bool = report_hess
        self.history_beats_optimizer: bool = history_beats_optimizer

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    @staticmethod
    def assert_instance(
        maybe_options: Union['OptimizeOptions', Dict],
    ) -> 'OptimizeOptions':
        """
        Return a valid options object.

        Parameters
        ----------
        maybe_options: OptimizeOptions or dict
        """
        if isinstance(maybe_options, OptimizeOptions):
            return maybe_options
        options = OptimizeOptions(**maybe_options)
        return options



import numbers
from typing import Dict, Tuple

import numpy as np

from ..problem import Problem
from .metropolis import MetropolisSampler


class AdaptiveMetropolisSampler(MetropolisSampler):
    """Metropolis-Hastings sampler with adaptive proposal covariance."""

    def __init__(self, options: Dict = None):
        super().__init__(options)
        self._cov = None
        self._mean_hist = None
        self._cov_hist = None
        self._cov_scale = None

    @classmethod
    def default_options(cls):
        """Return the default options for the sampler."""
        return {
            # controls adaptation degeneration velocity of the proposals
            # in [0, 1], with 0 -> no adaptation, i.e. classical
            # Metropolis-Hastings
            'decay_constant': 0.51,
            # number of samples before adaptation decreases significantly.
            # a higher value reduces the impact of early adaptation
            'threshold_sample': 1,
            # regularization factor for ill-conditioned cov matrices of
            # the adapted proposal density. regularization might happen if the
            # eigenvalues of the cov matrix strongly differ in order
            # of magnitude. in this case, the algorithm adds a small
            # diag matrix to the cov matrix with elements of this factor
            'reg_factor': 1e-6,
            # initial covariance matrix. defaults to a unit matrix
            'cov0': None,
            # target acceptance rate
            'target_acceptance_rate': 0.234,
            # show progress
            'show_progress': True,
        }

    def initialize(self, problem: Problem, x0: np.ndarray):
        """Initialize the sampler."""
        super().initialize(problem, x0)

        if self.options['cov0'] is not None:
            cov0 = self.options['cov0']
            if isinstance(cov0, numbers.Real):
                cov0 = float(cov0) * np.eye(len(x0))
        else:
            cov0 = np.eye(len(x0))
        self._cov = regularize_covariance(cov0, self.options['reg_factor'])
        self._mean_hist = self.trace_x[-1]
        self._cov_hist = self._cov
        self._cov_scale = 1.0

    def _propose_parameter(self, x: np.ndarray):
        x_new = np.random.multivariate_normal(x, self._cov)
        return x_new

    def _update_proposal(
        self, x: np.ndarray, lpost: float, log_p_acc: float, n_sample_cur: int
    ):
        # parse options
        decay_constant = self.options['decay_constant']
        threshold_sample = self.options['threshold_sample']
        reg_factor = self.options['reg_factor']
        target_acceptance_rate = self.options['target_acceptance_rate']

        # compute historical mean and covariance
        self._mean_hist, self._cov_hist = update_history_statistics(
            mean=self._mean_hist,
            cov=self._cov_hist,
            x_new=x,
            n_cur_sample=max(n_sample_cur + 1, threshold_sample),
            decay_constant=decay_constant,
        )

        # compute covariance scaling factor
        self._cov_scale *= np.exp(
            (np.exp(log_p_acc) - target_acceptance_rate)
            / np.power(n_sample_cur + 1, decay_constant)
        )

        # set proposal covariance
        # TODO check publication
        self._cov = self._cov_scale * self._cov_hist

        # regularize proposal covariance
        self._cov = regularize_covariance(cov=self._cov, reg_factor=reg_factor)


def update_history_statistics(
    mean: np.ndarray,
    cov: np.ndarray,
    x_new: np.ndarray,
    n_cur_sample: int,
    decay_constant: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Update sampling statistics.

    Parameters
    ----------
    mean:
        The estimated mean of the samples, that was calculated in the previous
        iteration.
    cov:
        The estimated covariance matrix of the sample, that was calculated in
        the previous iteration.
    x_new:
        Most recent sample.
    n_cur_sample:
        Current number of samples.
    decay_constant:
        Adaption decay, in [0, 1]. Higher values result in faster decays, such
        that later iterations influence the adaption more weakly.

    Returns
    -------
    mean, cov:
        The updated values for the estimated mean and the estimated covariance
        matrix of the sample.
    """
    update_rate = n_cur_sample ** (-decay_constant)

    mean = (1 - update_rate) * mean + update_rate * x_new

    dx = x_new - mean
    cov = (1 - update_rate) * cov + update_rate * dx.reshape(
        (-1, 1)
    ) @ dx.reshape((1, -1))

    return mean, cov


def regularize_covariance(cov: np.ndarray, reg_factor: float) -> np.ndarray:
    """
    Regularize the estimated covariance matrix of the sample.

    Useful if the estimated covariance matrix is ill-conditioned.
    Increments the diagonal a little to ensure positivity.

    Parameters
    ----------
    cov:
        Estimate of the covariance matrix of the sample.
    reg_factor:
        Regularization factor. Larger values result in stronger regularization.

    Returns
    -------
    cov:
        Regularized estimate of the covariance matrix of the sample.
    """
    eig = np.linalg.eigvals(cov)
    eig_min = min(eig)
    if eig_min <= 0:
        cov += (abs(eig_min) + reg_factor) * np.eye(cov.shape[0])
    return cov



"""
Prediction
==========

Generate predictions from simulations with specified parameter vectors, with
optional post-processing.
"""

from .amici_predictor import AmiciPredictor
from .task import PredictorTask



"""Engines without parallelization."""
from typing import List

from tqdm import tqdm

from .base import Engine
from .task import Task


class SingleCoreEngine(Engine):
    """
    Dummy engine for sequential execution on one core.

    Note that the objective itself may be multithreaded.
    """

    def __init__(self):
        super().__init__()

    def execute(self, tasks: List[Task], progress_bar: bool = True):
        """Execute all tasks in a simple for loop sequentially.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        results = []
        for task in tqdm(tasks, disable=not progress_bar):
            results.append(task.execute())

        return results



"""Enhanced Scatter Search.

See papers on ESS [EgeaBal2009]_ [EgeaMar2010]_, CESS [VillaverdeEge2012]_ and
saCeSS [PenasGon2017]_.

References
==========

.. [EgeaBal2009] 'Dynamic Optimization of Nonlinear Processes with an Enhanced
   Scatter Search Method', Jose A. Egea, Eva Balsa-Canto,
   María-Sonia G. García, and Julio R. Banga, Ind. Eng. Chem. Res.
   2009, 48, 9, 4388–4401. https://doi.org/10.1021/ie801717t

.. [EgeaMar2010] 'An evolutionary method for complex-process optimization',
   Jose A. Egea, Rafael Martí, Julio R. Banga, Computers & Operations Research,
   2010, 37, 2, 315-324. https://doi.org/10.1016/j.cor.2009.05.003

.. [VillaverdeEge2012] 'A cooperative strategy for parameter estimation in
   large scale systems biology models', Villaverde, A.F., Egea, J.A. & Banga,
   J.R. BMC Syst Biol 2012, 6, 75. https://doi.org/10.1186/1752-0509-6-75

.. [PenasGon2017] 'Parameter estimation in large-scale systems biology models:
   a parallel and self-adaptive cooperative strategy', David R. Penas,
   Patricia González, Jose A. Egea, Ramón Doallo and Julio R. Banga,
   BMC Bioinformatics 2017, 18, 52. https://doi.org/10.1186/s12859-016-1452-4
"""
import enum
import logging
import time
from typing import List, Optional, Tuple

import numpy as np

import pypesto.optimize
from pypesto import OptimizerResult, Problem
from pypesto.startpoint import StartpointMethod

from .function_evaluator import FunctionEvaluator
from .refset import RefSet

logger = logging.getLogger(__name__)

__all__ = ['ESSOptimizer', 'ESSExitFlag']


class ESSExitFlag(int, enum.Enum):
    """Exit flags used by :class:`ESSOptimizer`."""

    # ESS did not run/finish yet
    DID_NOT_RUN = 0
    # Exited after reaching maximum number of iterations
    MAX_ITER = -1
    # Exited after exhausting function evaluation budget
    MAX_EVAL = -2
    # Exited after exhausting wall-time budget
    MAX_TIME = -3


class ESSOptimizer:
    """Enhanced Scatter Search (ESS) global optimization.

    .. note: Does not implement any constraint handling beyond box constraints
    """

    def __init__(
        self,
        *,
        max_iter: int,
        local_n1: int,
        local_n2: int,
        dim_refset: int,
        balance: float,
        local_optimizer: 'pypesto.optimize.Optimizer' = None,
        max_eval=np.inf,
        n_diverse: int = None,
        n_threads=1,
        max_walltime_s=None,
    ):
        """Construct new ESS instance.

        For plausible values of hyperparameters, see VillaverdeEge2012.

        Parameters
        ----------
        dim_refset:
            Size of the ReferenceSet. Note that in every iteration at least
            ``dim_refset**2 - dim_refset`` function evaluations will occur.
        max_iter:
            Maximum number of ESS iterations.
        local_n1:
            Minimum number of iterations before first local search.
        local_n2:
            Minimum number of iterations between consecutive local
            searches. Maximally one local search per performed in each
            iteration.
        local_optimizer:
            Local optimizer for refinement, or ``None`` to skip local searches.
        n_diverse:
            Number of samples to choose from to construct the initial RefSet
        max_eval:
            Maximum number of objective functions allowed. This criterion is
            only checked once per iteration, not after every objective
            evaluation, so the actual number of function evaluations may exceed
            this value.
        max_walltime_s:
            Maximum walltime in seconds. Will only be checked between local
            optimizations and other simulations, and thus, may be exceeded by
            the duration of a local search.
        balance:
            Quality vs diversity balancing factor [0, 1];
            0 = only quality; 1 = only diversity
        """
        # Hyperparameters
        self.local_n1: int = local_n1
        self.local_n2: int = local_n2
        self.max_iter: int = max_iter
        self.max_eval: int = max_eval
        self.dim_refset: int = dim_refset
        self.local_optimizer: Optional[
            'pypesto.optimize.Optimizer'
        ] = local_optimizer
        self.n_diverse: int = n_diverse
        self.n_threads: int = n_threads
        self.balance: float = balance
        # After how many iterations a stagnated solution is to be replaced by
        #  a random one. Default value taken from [EgeaMar2010]_
        self.n_change: int = 20
        # Only perform local search from best solution
        self.local_only_best_sol: bool = False
        self.max_walltime_s = max_walltime_s
        self._initialize()

    def _initialize(self):
        """(Re-)Initialize."""
        # RefSet
        self.refset: Optional[RefSet] = None
        # Overall best parameters found so far
        self.x_best: Optional[np.array] = None
        # Overall best function value found so far
        self.fx_best: float = np.inf
        # Final parameters from local searches
        self.local_solutions: List[np.array] = []
        # Index of current iteration
        self.n_iter: int = 0
        # ESS iteration at which the last local search took place
        self.last_local_search_niter: int = 0
        # Whether self.x_best has changed in the current iteration
        self.x_best_has_changed: bool = False
        self.exit_flag: ESSExitFlag = ESSExitFlag.DID_NOT_RUN
        self.evaluator: Optional[FunctionEvaluator] = None
        self.starttime: Optional[float] = None

    def minimize(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod,
        refset: Optional[RefSet] = None,
    ) -> pypesto.Result:
        """Minimize the given objective.

        Parameters
        ----------
        problem:
            Problem to run ESS on.
        startpoint_method:
            Method for choosing starting points.
        refset:
            The initial RefSet or ``None`` to auto-generate.
        """
        self._initialize()
        self.starttime = time.time()

        # generate initial RefSet if not provided
        if refset is None:
            if self.n_diverse is None:
                # [EgeaMar2010]_ 2.1
                self.n_diverse = 10 * problem.dim

            self.evaluator = FunctionEvaluator(
                problem=problem,
                startpoint_method=startpoint_method,
                n_threads=self.n_threads,
            )
            self.refset = RefSet(dim=self.dim_refset, evaluator=self.evaluator)
            # Initial RefSet generation
            self.refset.initialize_random(n_diverse=self.n_diverse)
            refset = self.refset
        else:
            self.refset = refset

        self.evaluator = refset.evaluator
        self.x_best = np.full(
            shape=(self.evaluator.problem.dim,), fill_value=np.nan
        )
        # initialize global best from initial refset
        for x, fx in zip(refset.x, refset.fx):
            self._maybe_update_global_best(x, fx)

        # [PenasGon2017]_ Algorithm 1
        while self._keep_going():
            self.x_best_has_changed = False

            refset.sort()
            self._report_iteration()
            refset.prune_too_close()

            # Apply combination method to update the RefSet
            x_best_children, fx_best_children = self._combine_solutions()

            # Go-beyond strategy to further improve the new combinations
            self._go_beyond(x_best_children, fx_best_children)

            # Maybe perform a local search
            if self.local_optimizer is not None:
                self._do_local_search(x_best_children, fx_best_children)

            # Replace RefSet members by best children where an improvement
            #  was made. replace stuck members by random points.
            for i in range(refset.dim):
                if fx_best_children[i] < refset.fx[i]:
                    refset.update(i, x_best_children[i], fx_best_children[i])
                else:
                    refset.n_stuck[i] += 1
                    if refset.n_stuck[i] > self.n_change:
                        refset.replace_by_random(i)

            self.n_iter += 1

        self._report_final()
        return self._create_result()

    def _create_result(self) -> pypesto.Result:
        """Create the result object.

        Currently, this returns the overall best value and the final RefSet.
        """
        common_result_fields = {
            'exitflag': self.exit_flag,
            # meaningful? this is the overall time, and identical for all
            #  reported points
            'time': time.time() - self.starttime,
            'n_fval': self.evaluator.n_eval,
            'optimizer': str(self),
        }
        i_result = 0
        result = pypesto.Result(problem=self.evaluator.problem)

        # save global best
        optimizer_result = pypesto.OptimizerResult(
            id=str(i_result),
            x=self.x_best,
            fval=self.fx_best,
            message="Global best",
            **common_result_fields,
        )
        # TODO DW: Create a single History with the global best?
        result.optimize_result.append(optimizer_result)

        # save refset
        for i in range(self.refset.dim):
            i_result += 1
            result.optimize_result.append(
                pypesto.OptimizerResult(
                    id=str(i_result),
                    x=self.refset.x[i],
                    fval=self.refset.fx[i],
                    message=f"RefSet[{i}]",
                    **common_result_fields,
                )
            )

        # TODO DW: also save local solutions?
        #  (need to track fvals or re-evaluate)

        return result

    def _keep_going(self) -> bool:
        """Check exit criteria.

        Returns
        -------
        ``True`` if not of the exit criteria is met, ``False`` otherwise.
        """
        # TODO DW which further stopping criteria: gtol, fatol, frtol?

        if self.n_iter >= self.max_iter:
            self.exit_flag = ESSExitFlag.MAX_ITER
            return False

        if self.evaluator.n_eval >= self.max_eval:
            self.exit_flag = ESSExitFlag.MAX_EVAL
            return False

        if (
            self.max_walltime_s is not None
            and time.time() - self.starttime > self.max_walltime_s
        ):
            self.exit_flag = ESSExitFlag.MAX_TIME
            return False

        return True

    def _combine_solutions(self) -> Tuple[np.array, np.array]:
        """Combine solutions and evaluate.

        Creates the next generation from the RefSet by pair-wise combinations
        of all RefSet members. Creates ``RefSet.dim ** 2 - RefSet.dim`` new
        parameter vectors, tests them, and keeps the best child of each parent.

        Returns
        -------
        y:
            Contains the best of all pairwise combinations of all RefSet
            members, for each RefSet members.
        fy:
            The corresponding objective values.
        """
        y = np.zeros(shape=(self.refset.dim, self.evaluator.problem.dim))
        fy = np.full(shape=self.refset.dim, fill_value=np.inf)
        for i in range(self.refset.dim):
            xs_new = np.vstack(
                tuple(
                    self._combine(i, j)
                    for j in range(self.refset.dim)
                    if i != j
                ),
            )
            fxs_new = self.evaluator.multiple(xs_new)
            best_idx = np.argmin(fxs_new)
            fy[i] = fxs_new[best_idx]
            y[i] = xs_new[best_idx]
        return y, fy

    def _combine(self, i, j) -> np.array:
        """Combine RefSet members ``i`` and ``j``.

        Samples a new point from a biased hyper-rectangle derived from the
        given parents, favoring the direction of the better parent.

        Assumes that the RefSet is sorted by quality.

        See [EgeaBal2009]_ Section 3.2 for details.

        Parameters
        ----------
        i:
            Index of first RefSet member for recombination
        j:
            Index of second RefSet member for recombination

        Returns
        -------
        A new parameter vector.
        """
        if i == j:
            raise ValueError("i == j")
        x = self.refset.x

        d = x[j] - x[i]
        alpha = np.sign(j - i)
        beta = (np.abs(j - i) - 1) / (self.refset.dim - 2)
        c1 = x[i] - d * (1 + alpha * beta)
        c2 = x[i] - d * (1 - alpha * beta)

        # this will not always yield admissible points -> clip to bounds
        ub, lb = self.evaluator.problem.ub, self.evaluator.problem.lb
        c1 = np.fmax(np.fmin(c1, ub), lb)
        c2 = np.fmax(np.fmin(c2, ub), lb)

        return np.random.uniform(
            low=c1, high=c2, size=self.evaluator.problem.dim
        )

    def _do_local_search(
        self, x_best_children: np.array, fx_best_children: np.array
    ) -> None:
        """
        Perform a local search to refine the next generation.

        See [PenasGon2017]_ Algorithm 2.
        """
        if self.local_only_best_sol and self.x_best_has_changed:
            logger.debug("Local search only from best point.")
            local_search_x0 = self.x_best
            local_search_fx0 = self.fx_best
        # first local search?
        elif not self.local_solutions and self.n_iter >= self.local_n1:
            logger.debug(
                "First local search from best point due to "
                f"local_n1={self.local_n1}."
            )
            local_search_x0 = self.x_best
            local_search_fx0 = self.fx_best
        elif (
            self.local_solutions
            and self.n_iter - self.last_local_search_niter >= self.local_n2
        ):
            quality_order = np.argsort(fx_best_children)
            # compute minimal distance between the best children and all local
            #  optima found so far
            min_distances = np.array(
                np.min(
                    np.linalg.norm(y_i - local_solution)
                    for local_solution in self.local_solutions
                )
                for y_i in x_best_children
            )
            # sort by furthest distance to existing local optima
            diversity_order = np.argsort(min_distances)[::-1]
            # compute priority, balancing quality and diversity
            #  (smaller value = higher priority)
            priority = (
                1 - self.balance
            ) * quality_order + self.balance * diversity_order
            chosen_child_idx = np.argmin(priority)
            local_search_x0 = x_best_children[chosen_child_idx]
            local_search_fx0 = fx_best_children[chosen_child_idx]
        else:
            return

        # actual local search
        # TODO DW: try alternatives if it fails on initial point?
        optimizer_result: OptimizerResult = self.local_optimizer.minimize(
            problem=self.evaluator.problem,
            x0=local_search_x0,
            id="0",
        )
        # add function evaluations during local search to our function
        #  evaluation counter (NOTE: depending on the setup, we might neglect
        #  gradient evaluations).
        self.evaluator.n_eval += optimizer_result.n_fval
        self.evaluator.n_eval_round += optimizer_result.n_fval

        logger.info(
            f"Local search: {local_search_fx0} -> {optimizer_result.fval} "
            f" took {optimizer_result.time:.3g}s, finished with "
            f"{optimizer_result.exitflag}: {optimizer_result.message}"
        )
        self.local_solutions.append(optimizer_result.x)

        self._maybe_update_global_best(
            optimizer_result.x, optimizer_result.fval
        )
        self.last_local_search_niter = self.n_iter
        self.evaluator.reset_round_counter()

    def _maybe_update_global_best(self, x, fx):
        """Update the global best value if the provided value is better."""
        if fx < self.fx_best:
            self.x_best = x[:]
            self.fx_best = fx
            self.x_best_has_changed = True

    def _go_beyond(self, x_best_children, fx_best_children):
        """Apply go-beyond strategy.

        If a child is better than its parent, intensify search in that
        direction until no further improvement is made.

        See [Egea2009]_ algorithm 1 + section 3.4
        """
        for i in range(self.refset.dim):
            if fx_best_children[i] >= self.refset.fx[i]:
                continue

            # offspring is better than parent
            x_parent = self.refset.x[i]
            fx_parent = self.refset.fx[i]
            x_child = x_best_children[i]
            fx_child = fx_best_children[i]
            improvement = 1
            # Multiplier used in determining the hyper-rectangle from which to
            # sample children. Will be increased in case of 2 consecutive
            # improvements.
            # (corresponds to 1/\Lambda in [Egea2009]_ algorithm 1)
            go_beyond_factor = 1
            while fx_child < fx_parent:
                # update best child
                x_best_children[i] = x_child
                fx_best_children[i] = fx_child

                # create new solution, child becomes parent
                # hyper-rectangle for sampling child
                box_lb = x_child - (x_parent - x_child) * go_beyond_factor
                box_ub = x_child
                # clip to bounds
                ub, lb = self.evaluator.problem.ub, self.evaluator.problem.lb
                box_lb = np.fmax(np.fmin(box_lb, ub), lb)
                box_ub = np.fmax(np.fmin(box_ub, ub), lb)
                # sample parameters
                x_new = np.random.uniform(low=box_lb, high=box_ub)
                x_parent = x_child
                fx_parent = fx_child
                x_child = x_new
                fx_child = self.evaluator.single(x_child)

                improvement += 1
                if improvement == 2:
                    go_beyond_factor *= 2
                    improvement = 0

            # update overall best?
            self._maybe_update_global_best(
                x_best_children[i], fx_best_children[i]
            )

    def _report_iteration(self):
        """Log the current iteration."""
        if self.n_iter == 0:
            logger.info("iter | best | nf | refset         |")

        with np.printoptions(
            edgeitems=30,
            linewidth=100000,
            formatter={"float": lambda x: "%.3g" % x},
        ):
            logger.info(
                f"{self.n_iter:4} | {self.fx_best:+.2E} | "
                f"{self.evaluator.n_eval} "
                f"| {self.refset.fx} | {len(self.local_solutions)}"
            )

    def _report_final(self):
        """Log scatter search summary."""
        with np.printoptions(
            edgeitems=30,
            linewidth=100000,
            formatter={"float": lambda x: "%.3g" % x},
        ):
            logger.info(
                f"-- {self.exit_flag!r} "
                f"Stopping after {self.n_iter} iterations. "
                f"Num local solutions: {len(self.local_solutions)}."
            )
            logger.info(f"Final refset: {np.sort(self.refset.fx)} ")
        logger.info(f"Best fval {self.fx_best}")



import logging
from typing import Any, Callable, List

import numpy as np

from ..engine import Task

logger = logging.getLogger(__name__)


class EnsembleTask(Task):
    """Apply a method to each vector in a set of parameter vectors.

    Attributes
    ----------
    method:
        The method.
    vectors:
        The parameter vectors, with shape `(n_parameters, n_vectors)`.
    id:
        The task ID.
    """

    def __init__(
        self,
        method: Callable,
        vectors: np.ndarray,
        id: str,
    ):
        super().__init__()
        self.method = method
        self.vectors = vectors
        self.id = id

    def execute(self) -> List[Any]:
        """Execute the task."""
        logger.info(f"Executing task {self.id}.")
        results = []
        for index in range(self.vectors.shape[1]):
            results.append(self.method(self.vectors[:, index]))
        return results



"""AdaptiveParallelTemperingSampler class."""
from typing import Dict, Sequence

import numpy as np

from .parallel_tempering import ParallelTemperingSampler


class AdaptiveParallelTemperingSampler(ParallelTemperingSampler):
    """Parallel tempering sampler with adaptive temperature adaptation."""

    @classmethod
    def default_options(cls) -> Dict:
        """Get default options for sampler."""
        options = super().default_options()
        # scaling factor for temperature adaptation
        options['eta'] = 100
        # controls the adaptation degeneration velocity of the temperature
        # adaption.
        options['nu'] = 1e3

        return options

    def adjust_betas(self, i_sample: int, swapped: Sequence[bool]):
        """Update temperatures as in Vousden2016."""
        if len(self.betas) == 1:
            return

        # parameters
        nu = self.options['nu']
        eta = self.options['eta']
        betas = self.betas

        # booleans to integer array
        swapped = np.array([int(swap) for swap in swapped])

        # update betas
        kappa = nu / (i_sample + 1 + nu) / eta
        ds = kappa * (swapped[:-1] - swapped[1:])
        dtemp = np.diff(1.0 / betas[:-1])
        dtemp = dtemp * np.exp(ds)
        betas[:-1] = 1 / np.cumsum(np.insert(dtemp, obj=0, values=1.0))

        # fill in
        self.betas = betas



"""Abstract Task class."""
import abc


class Task(abc.ABC):
    """
    Abstract Task class.

    A task is one of a list of independent execution tasks that are
    submitted to the execution engine to be executed using the execute()
    method, commonly in parallel.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def execute(self):
        """Execute the task and return its results."""



"""Example problems for tests / demos."""

import pandas as pd

from pypesto.C import InnerParameterType


def get_Boehm_JProteomeRes2014_hierarchical_petab() -> 'petab.Problem':  # noqa: F821
    """
    Get Boehm_JProteomeRes2014 problem with scaled/offset observables.

    Creates a modified version of the Boehm_JProteomeRes2014 benchmark problem,
    suitable for hierarchical optimization.
    """
    import petab
    from benchmark_models_petab import get_problem

    from pypesto.hierarchical.problem import PARAMETER_TYPE

    petab_problem = get_problem("Boehm_JProteomeRes2014")
    # Add scaling and offset parameters
    petab_problem.observable_df[petab.OBSERVABLE_FORMULA] = [
        f"observableParameter2_{obs_id} + observableParameter1_{obs_id} "
        f"* {obs_formula}"
        for obs_id, obs_formula in zip(
            petab_problem.observable_df.index,
            petab_problem.observable_df[petab.OBSERVABLE_FORMULA],
        )
    ]
    # Set scaling and offset parameters for measurements
    if not (
        petab_problem.measurement_df[petab.OBSERVABLE_PARAMETERS].isna().all()
    ):
        raise AssertionError()
    petab_problem.measurement_df[petab.OBSERVABLE_PARAMETERS] = [
        f"scaling_{obs_id};offset_{obs_id}"
        for obs_id in petab_problem.measurement_df[petab.OBSERVABLE_ID]
    ]
    # Add output parameters to parameter table
    extra_parameters = [
        {
            petab.PARAMETER_ID: par_id,
            petab.PARAMETER_SCALE: petab.LIN,
            petab.LOWER_BOUND: -100,
            petab.UPPER_BOUND: 100,
            petab.NOMINAL_VALUE: 0,
            petab.ESTIMATE: 1,
        }
        for par_id in (
            'offset_pSTAT5A_rel',
            'offset_pSTAT5B_rel',
            'offset_rSTAT5A_rel',
        )
    ]

    extra_parameters.extend(
        {
            petab.PARAMETER_ID: par_id,
            petab.PARAMETER_SCALE: petab.LIN,
            petab.LOWER_BOUND: 1e-5,
            petab.UPPER_BOUND: 1e5,
            petab.NOMINAL_VALUE: nominal_value,
            petab.ESTIMATE: 1,
        }
        for par_id, nominal_value in zip(
            (
                'scaling_pSTAT5A_rel',
                'scaling_pSTAT5B_rel',
                'scaling_rSTAT5A_rel',
            ),
            (3.85261197844677, 6.59147818673419, 3.15271275648527),
        )
    )

    petab_problem.parameter_df = pd.concat(
        [
            petab_problem.parameter_df,
            petab.get_parameter_df(pd.DataFrame(extra_parameters)),
        ]
    )
    # Mark output parameters for hierarchical optimization
    petab_problem.parameter_df[PARAMETER_TYPE] = None
    for par_id in petab_problem.parameter_df.index:
        if par_id.startswith("offset_"):
            petab_problem.parameter_df.loc[
                par_id, PARAMETER_TYPE
            ] = InnerParameterType.OFFSET.value
        elif par_id.startswith("sd_"):
            petab_problem.parameter_df.loc[
                par_id, PARAMETER_TYPE
            ] = InnerParameterType.SIGMA.value
        elif par_id.startswith("scaling_"):
            petab_problem.parameter_df.loc[
                par_id, PARAMETER_TYPE
            ] = InnerParameterType.SCALING.value

    # log-scaling is not supported for inner parameters
    petab_problem.parameter_df.loc[
        petab_problem.parameter_df.index.str.startswith("sd_"),
        petab.PARAMETER_SCALE,
    ] = petab.LIN

    petab.lint_problem(petab_problem)

    return petab_problem


def get_Boehm_JProteomeRes2014_hierarchical_petab_corrected_bounds() -> 'petab.Problem':  # noqa: F821
    """
    See `get_Boehm_JProteomeRes2014_hierarchical_petab`.

    Adjusts this PEtab problem to have correct bounds for inner parameters.
    """
    from ..hierarchical.petab import correct_parameter_df_bounds

    petab_problem = get_Boehm_JProteomeRes2014_hierarchical_petab()
    petab_problem.parameter_df = correct_parameter_df_bounds(
        parameter_df=petab_problem.parameter_df
    )
    return petab_problem



from __future__ import annotations

import logging
import numbers
import warnings
from typing import TYPE_CHECKING, Dict, Sequence, Tuple, Union

import numpy as np

from ...C import (
    FVAL,
    GRAD,
    HESS,
    MODE_FUN,
    MODE_RES,
    RDATAS,
    RES,
    SRES,
    ModeType,
)
from ...logging import log_level_active

if TYPE_CHECKING:
    try:
        import amici
        from amici.parameter_mapping import (
            ParameterMapping,
            ParameterMappingForCondition,
        )
    except ImportError:
        ParameterMapping = ParameterMappingForCondition = None

AmiciModel = Union['amici.Model', 'amici.ModelPtr']
AmiciSolver = Union['amici.Solver', 'amici.SolverPtr']

logger = logging.getLogger(__name__)


def map_par_opt_to_par_sim(
    condition_map_sim_var: Dict[str, Union[float, str]],
    x_dct: Dict[str, float],
    amici_model: AmiciModel,
) -> np.ndarray:
    """
    Create simulation vector from optimization vector using the mapping.

    Parameters
    ----------
    condition_map_sim_var:
        Simulation to optimization parameter mapping.
    x_dct:
        The optimization parameters dict.
    amici_model:
        The amici model.

    Returns
    -------
    par_sim_vals:
        The simulation parameters vector corresponding to x under the
        specified mapping.
    """
    par_sim_vals = [
        condition_map_sim_var[par_id]
        for par_id in amici_model.getParameterIds()
    ]

    # iterate over simulation parameter indices
    for ix, val in enumerate(par_sim_vals):
        if not isinstance(val, numbers.Number):
            # value is optimization parameter id
            par_sim_vals[ix] = x_dct[val]

    # return the created simulation parameter vector
    return np.array(par_sim_vals)


def create_plist_from_par_opt_to_par_sim(mapping_par_opt_to_par_sim):
    """
    Create list of parameter indices for which sensitivity is to be computed.

    From the parameter mapping `mapping_par_opt_to_par_sim`, create the
    simulation plist according to the mapping `mapping`.

    Parameters
    ----------
    mapping_par_opt_to_par_sim: array-like of str
        len == n_par_sim, the entries are either numeric, or
        optimization parameter ids.

    Returns
    -------
    plist: array-like of float
        List of parameter indices for which the sensitivity needs to be
        computed
    """
    warnings.warn(
        "This function will be removed in future releases. ",
        DeprecationWarning,
    )
    plist = []

    # iterate over simulation parameter indices
    for j_par_sim, val in enumerate(mapping_par_opt_to_par_sim):
        if not isinstance(val, numbers.Number):
            plist.append(j_par_sim)

    # return the created simulation parameter vector
    return plist


def create_identity_parameter_mapping(
    amici_model: AmiciModel, n_conditions: int
) -> ParameterMapping:
    """Create a dummy identity parameter mapping table.

    This fills in only the dynamic parameters. Values for fixed parameters,
    both in preequilibration and simulation, are assumed to be provided
    correctly in model or edatas already.
    """
    import amici.parameter_mapping

    x_ids = list(amici_model.getParameterIds())
    x_scales = list(amici_model.getParameterScale())
    parameter_mapping = amici.parameter_mapping.ParameterMapping()
    for _ in range(n_conditions):
        condition_map_sim_var = {x_id: x_id for x_id in x_ids}
        condition_scale_map_sim_var = {
            x_id: amici.parameter_mapping.amici_to_petab_scale(x_scale)
            for x_id, x_scale in zip(x_ids, x_scales)
        }
        # assumes fixed parameters are filled in already
        mapping_for_condition = (
            amici.parameter_mapping.ParameterMappingForCondition(
                map_sim_var=condition_map_sim_var,
                scale_map_sim_var=condition_scale_map_sim_var,
            )
        )

        parameter_mapping.append(mapping_for_condition)
    return parameter_mapping


def par_index_slices(
    par_opt_ids: Sequence[str],
    par_sim_ids: Sequence[str],
    condition_map_sim_var: Dict[str, Union[float, str]],
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate numpy arrays for indexing based on `mapping_par_opt_to_par_sim`.

    Parameters
    ----------
    par_opt_ids:
        The optimization parameter ids. Needed for order.
    par_sim_ids:
        The simulation parameter ids. Needed for order.
    condition_map_sim_var:
        The simulation to optimization parameter mapping.

    Returns
    -------
    par_sim_slice:
        array of simulation parameter indices
    par_opt_slice:
        array of optimization parameter indices
    """
    # Create ID to index mapping for more efficient lookup than list.index
    par_opt_id_to_idx = {id_: idx for idx, id_ in enumerate(par_opt_ids)}
    par_sim_id_to_idx = {id_: idx for idx, id_ in enumerate(par_sim_ids)}

    # bool array indicating which simulation parameters map to any estimated
    #  parameters
    par_sim_maps_to_str = np.fromiter(
        (
            isinstance(condition_map_sim_var[par_id], str)
            for par_id in par_sim_ids
        ),
        dtype=bool,
        count=len(par_sim_ids),
    )
    # elsewhere, amici.ExpData.plist is set to compute only sensitivities
    #  w.r.t. estimated parameters. the parameter ordering is preserved there.
    #  therefore, the cumulative sum of mapped estimated parameters yields the
    #  index for AMICI-computed sensitivities for the respective parameter.
    cumsum_par_sim_maps_to_str = np.cumsum(par_sim_maps_to_str)

    zip_iterator = zip(
        *(
            (
                # sensitivity parameter index in AMICI simulation results
                cumsum_par_sim_maps_to_str[par_sim_id_to_idx[par_sim_id]] - 1,
                # corresponding optimization parameter index
                par_opt_id_to_idx[par_opt_id],
            )
            for par_sim_id, par_opt_id in condition_map_sim_var.items()
            if isinstance(par_opt_id, str) and par_opt_id in par_opt_id_to_idx
        )
    )
    par_sim_slice = np.fromiter(next(zip_iterator), dtype=int)
    par_opt_slice = np.fromiter(next(zip_iterator), dtype=int)
    return par_sim_slice, par_opt_slice


def add_sim_grad_to_opt_grad(
    par_opt_ids: Sequence[str],
    par_sim_ids: Sequence[str],
    condition_map_sim_var: Dict[str, Union[float, str]],
    sim_grad: np.ndarray,
    opt_grad: np.ndarray,
    coefficient: float = 1.0,
) -> None:
    """
    Sum simulation gradients to objective gradient.

    Uses the provided mapping `mapping_par_opt_to_par_sim` for summing up.

    Parameters
    ----------
    par_opt_ids:
        The optimization parameter ids. Needed for order.
    par_sim_ids:
        The simulation parameter ids. Needed for order.
    condition_map_sim_var:
        The simulation to optimization parameter mapping.
    sim_grad:
        Simulation gradient.
    opt_grad:
        The optimization gradient. To which sim_grad is added.
        Changed in-place.
    coefficient:
        Coefficient for sim_grad when adding to opt_grad.
    """
    par_sim_slice, par_opt_slice = par_index_slices(
        par_opt_ids, par_sim_ids, condition_map_sim_var
    )

    par_opt_slice_unique, unique_index = np.unique(
        par_opt_slice, return_index=True
    )
    opt_grad[par_opt_slice_unique] += (
        coefficient * sim_grad[par_sim_slice[unique_index]]
    )

    if par_opt_slice_unique.size < par_opt_slice.size:
        for idx in range(len(par_opt_slice)):
            if idx not in unique_index:
                opt_grad[par_opt_slice[idx]] += (
                    coefficient * sim_grad[par_sim_slice[idx]]
                )


def add_sim_hess_to_opt_hess(
    par_opt_ids: Sequence[str],
    par_sim_ids: Sequence[str],
    condition_map_sim_var: Dict[str, Union[float, str]],
    sim_hess: np.ndarray,
    opt_hess: np.ndarray,
    coefficient: float = 1.0,
) -> None:
    """
    Sum simulation hessians to objective hessian.

    Parameters
    ----------
    Same as for add_sim_grad_to_opt_grad, replacing the gradients by hessians.
    """
    par_sim_slice, par_opt_slice = par_index_slices(
        par_opt_ids, par_sim_ids, condition_map_sim_var
    )

    par_opt_slice_unique, unique_index = np.unique(
        par_opt_slice, return_index=True
    )

    non_unique_indices = [
        idx for idx in range(len(par_opt_slice)) if idx not in unique_index
    ]

    opt_hess[np.ix_(par_opt_slice_unique, par_opt_slice_unique)] += (
        coefficient
        * sim_hess[
            np.ix_(par_sim_slice[unique_index], par_sim_slice[unique_index])
        ]
    )

    if par_opt_slice_unique.size < par_opt_slice.size:
        for idx in non_unique_indices:
            opt_hess[par_opt_slice[idx], par_opt_slice_unique] += (
                coefficient
                * sim_hess[par_sim_slice[idx], par_sim_slice[unique_index]]
            )
            opt_hess[par_opt_slice_unique, par_opt_slice[idx]] += (
                coefficient
                * sim_hess[par_sim_slice[unique_index], par_sim_slice[idx]]
            )
            for jdx in non_unique_indices:
                opt_hess[par_opt_slice[idx], par_opt_slice[jdx]] += (
                    coefficient
                    * sim_hess[par_sim_slice[idx], par_sim_slice[jdx]]
                )


def sim_sres_to_opt_sres(
    par_opt_ids: Sequence[str],
    par_sim_ids: Sequence[str],
    condition_map_sim_var: Dict[str, Union[float, str]],
    sim_sres: np.ndarray,
    coefficient: float = 1.0,
) -> np.ndarray:
    """

    Sum simulation residual sensitivities to objective residual sensitivities.

    Parameters
    ----------
    Mostly the same as for add_sim_grad_to_opt_grad, replacing the gradients by
    residual sensitivities.
    """
    opt_sres = np.zeros((sim_sres.shape[0], len(par_opt_ids)))

    par_sim_slice, par_opt_slice = par_index_slices(
        par_opt_ids, par_sim_ids, condition_map_sim_var
    )

    par_opt_slice_unique, unique_index = np.unique(
        par_opt_slice, return_index=True
    )
    opt_sres[:, par_opt_slice_unique] += (
        coefficient * sim_sres[:, par_sim_slice[unique_index]]
    )

    if par_opt_slice_unique.size < par_opt_slice.size:
        for idx in range(len(par_opt_slice)):
            if idx not in unique_index:
                opt_sres[:, par_opt_slice[idx]] += (
                    coefficient * sim_sres[:, par_sim_slice[idx]]
                )

    return opt_sres


def log_simulation(data_ix, rdata) -> None:
    """Log the simulation results."""
    logger.debug(f"=== DATASET {data_ix} ===")
    logger.debug(f"status: {rdata['status']}")
    logger.debug(f"llh: {rdata['llh']}")

    t_steadystate = 't_steadystate'
    if t_steadystate in rdata and rdata[t_steadystate] != np.nan:
        logger.debug(f"t_steadystate: {rdata[t_steadystate]}")

    if log_level_active(logger, logging.DEBUG):
        logger.debug(f"res: {rdata['res']}")


def get_error_output(
    amici_model: AmiciModel,
    edatas: Sequence['amici.ExpData'],
    rdatas: Sequence['amici.ReturnData'],
    sensi_orders: Tuple[int, ...],
    mode: ModeType,
    dim: int,
) -> dict:
    """Get default output upon error.

    Returns values indicative of an error, that is with nan entries in all
    vectors, and a function value, i.e. nllh, of `np.inf`.
    """
    if not amici_model.nt():
        nt = sum(data.nt() for data in edatas)
    else:
        nt = sum(
            data.nt() if data.nt() else amici_model.nt() for data in edatas
        )
    n_res = nt * amici_model.nytrue
    if amici_model.getAddSigmaResiduals():
        n_res *= 2

    nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
        sensi_orders, mode, dim, True
    )
    if res is not None:
        res = np.nan * np.ones(n_res)
    if sres is not None:
        sres = np.nan * np.ones([n_res, dim])

    ret = {
        FVAL: nllh,
        GRAD: snllh,
        HESS: s2nllh,
        RES: res,
        SRES: sres,
        RDATAS: rdatas,
    }
    return filter_return_dict(ret)


def init_return_values(
    sensi_orders: Tuple[int, ...],
    mode: ModeType,
    dim: int,
    error: bool = False,
):
    """Initialize return values."""
    if error:
        fval = np.inf
        sval = np.nan
    else:
        fval = sval = 0.0

    nllh = fval
    snllh = None
    s2nllh = None
    if mode == MODE_FUN:
        if 1 in sensi_orders:
            snllh = sval * np.ones(dim)
        if 2 in sensi_orders:
            s2nllh = sval * np.ones([dim, dim])

    chi2 = None
    res = None
    sres = None
    if mode == MODE_RES:
        if 0 in sensi_orders:
            chi2 = fval
            res = np.zeros([0])
        if 1 in sensi_orders:
            sres = np.zeros([0, dim])

    return nllh, snllh, s2nllh, chi2, res, sres


def filter_return_dict(ret) -> dict:
    """Filter return dict for non-None values."""
    return {key: val for key, val in ret.items() if val is not None}



from numbers import Real
from typing import Dict, Iterable, List, Optional, Sequence, Tuple, Union

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

from pypesto.util import delete_nan_inf

from ..result import Result
from .clust_color import assign_colors, assign_colors_for_list
from .misc import process_result_list, process_start_indices


def optimization_run_properties_one_plot(
    results: Result,
    properties_to_plot: Optional[List[str]] = None,
    size: Tuple[float, float] = (18.5, 10.5),
    start_indices: Optional[Union[int, Iterable[int]]] = None,
    colors: Optional[Union[List[float], List[List[float]]]] = None,
    legends: Optional[Union[str, List[str]]] = None,
    plot_type: str = 'line',
) -> matplotlib.axes.Axes:
    """
    Plot stats for allproperties specified in properties_to_plot on one plot.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    properties_to_plot:
        Optimization run properties that should be plotted
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    colors:
        List of RGBA colors (one color per property in properties_to_plot),
        or single RGBA color. If not set and one result, clustering is done
        and colors are assigned automatically
    legends:
        Labels, one label per optimization property
    plot_type:
        Specifies plot type. Possible values: 'line' and 'hist'

    Returns
    -------
    ax:
        The plot axes.

    Examples
    --------
    optimization_properties_per_multistart(
        result1,
        properties_to_plot=['time'],
        colors=[.5, .9, .9, .3])

    optimization_properties_per_multistart(
        result1,
        properties_to_plot=['time', 'n_grad'],
        colors=[[.5, .9, .9, .3], [.2, .1, .9, .5]])
    """
    if properties_to_plot is None:
        properties_to_plot = [
            'time',
            'n_fval',
            'n_grad',
            'n_hess',
            'n_res',
            'n_sres',
        ]

    if colors is None:
        colors = assign_colors_for_list(len(properties_to_plot))
    elif len(colors) == 4 and isinstance(colors[0], Real):
        colors = [colors]

    if len(colors) != len(properties_to_plot):
        raise ValueError(
            'Number of RGBA colors should be the same as number '
            'of optimization properties to plot'
        )

    if legends is None:
        legends = properties_to_plot
    elif not isinstance(legends, list):
        legends = [legends]

    if len(legends) != len(properties_to_plot):
        raise ValueError(
            'Number of legends should be the same as number of '
            'optimization properties to plot'
        )

    ax = plt.subplots()[1]
    fig = plt.gcf()
    fig.set_size_inches(*size)

    for idx, prop_name in enumerate(properties_to_plot):
        optimization_run_property_per_multistart(
            results,
            prop_name,
            ax,
            size,
            start_indices,
            colors[idx],
            legends[idx],
            plot_type,
        )

    ax.set_ylabel("property value")
    ax.set_title("Optimization properties per optimization run")
    return ax


def optimization_run_properties_per_multistart(
    results: Union[Result, Sequence[Result]],
    properties_to_plot: Optional[List[str]] = None,
    size: Tuple[float, float] = (18.5, 10.5),
    start_indices: Optional[Union[int, Iterable[int]]] = None,
    colors: Optional[Union[List[float], List[List[float]]]] = None,
    legends: Optional[Union[str, List[str]]] = None,
    plot_type: str = 'line',
) -> Dict[str, plt.Subplot]:
    """
    One plot per optimization property in properties_to_plot.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    properties_to_plot:
        Optimization run properties that should be plotted
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    colors:
        List of RGBA colors (one color per result in results),
        or single RGBA color. If not set and one result, clustering is done
        and colors are assigned automatically
    legends:
        Labels for line plots, one label per result object
    plot_type:
        Specifies plot type. Possible values: 'line' and 'hist'

    Returns
    -------
    ax:
    The plot axes.

    Examples
    --------
    optimization_properties_per_multistart(
        result1,
        properties_to_plot=['time'],
        colors=[.5, .9, .9, .3])

    optimization_properties_per_multistart(
        [result1, result2],
        properties_to_plot=['time'],
        colors=[[.5, .9, .9, .3], [.2, .1, .9, .5]])

    optimization_properties_per_multistart(
        result1,
        properties_to_plot=['time', 'n_grad'],
        colors=[.5, .9, .9, .3])

    optimization_properties_per_multistart(
        [result1, result2], properties_to_plot=['time', 'n_fval'],
        colors=[[.5, .9, .9, .3], [.2, .1, .9, .5]])
    """
    if properties_to_plot is None:
        properties_to_plot = [
            'time',
            'n_fval',
            'n_grad',
            'n_hess',
            'n_res',
            'n_sres',
        ]

    num_subplot = len(properties_to_plot)
    # compute, how many rows and columns we need for the subplots
    num_row = int(np.round(np.sqrt(num_subplot)))
    num_col = int(np.ceil(num_subplot / num_row))
    fig, axes = plt.subplots(num_row, num_col, squeeze=False)
    fig.set_size_inches(*size)

    for ax in axes.flat[num_subplot:]:
        ax.remove()
    axes = dict(zip(range(num_subplot), axes.flat))
    for idx, prop_name in enumerate(properties_to_plot):
        ax = axes[idx]
        optimization_run_property_per_multistart(
            results,
            prop_name,
            ax,
            size,
            start_indices,
            colors,
            legends,
            plot_type,
        )
    return axes


def optimization_run_property_per_multistart(
    results: Union[Result, Sequence[Result]],
    opt_run_property: str,
    axes: Optional[matplotlib.axes.Axes] = None,
    size: Tuple[float, float] = (18.5, 10.5),
    start_indices: Optional[Union[int, Iterable[int]]] = None,
    colors: Optional[Union[List[float], List[List[float]]]] = None,
    legends: Optional[Union[str, List[str]]] = None,
    plot_type: str = 'line',
) -> matplotlib.axes.Axes:
    """
    Plot stats for an optimization run property specified by opt_run_property.

    It is possible to plot a histogram or a line plot. In a line plot,
    on the x axis are the numbers of the multistarts, where the multistarts are
    ordered with respect to a function value. On the y axis of the line plot
    the value of the corresponding parameter for each multistart is displayed.

    Parameters
    ----------
    opt_run_property:
        optimization run property to plot.
        One of the 'time', 'n_fval', 'n_grad', 'n_hess', 'n_res', 'n_sres'
    results:
        Optimization result obtained by 'optimize.py' or list of those
    axes:
        Axes object to use
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    colors:
        List of RGBA colors (one color per result in results),
        or single RGBA color. If not set and one result, clustering is done
        and colors are assigned automatically
    legends:
        Labels for line plots, one label per result object
    plot_type:
        Specifies plot type. Possible values: 'line', 'hist', 'both'

    Returns
    -------
    axes:
        The plot axes.
    """
    supported_properties = {
        'time': 'Wall-clock time (seconds)',
        'n_fval': 'Number of function evaluations',
        'n_grad': 'Number of gradient evaluations',
        'n_hess': 'Number of Hessian evaluations',
        'n_res': 'Number of residuals evaluations',
        'n_sres': 'Number of residual sensitivity evaluations',
    }

    if opt_run_property not in supported_properties:
        raise ValueError(
            "Wrong value of opt_run_property. Only the following "
            "values are allowed: 'time', 'n_fval', 'n_grad', "
            "'n_hess', 'n_res', 'n_sres'"
        )

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    # axes
    if axes is None:
        ncols = 2 if plot_type == 'both' else 1
        fig, axes = plt.subplots(1, ncols)
        fig.set_size_inches(*size)
        fig.suptitle(
            f'{supported_properties[opt_run_property]} per optimizer run'
        )
    else:
        axes.set_title(
            f'{supported_properties[opt_run_property]} per optimizer run'
        )

    # loop over results
    for j, result in enumerate(results):
        if plot_type == 'both':
            axes[0] = stats_lowlevel(
                result,
                opt_run_property,
                supported_properties[opt_run_property],
                axes[0],
                start_indices,
                colors[j],
                legends[j],
            )

            axes[1] = stats_lowlevel(
                result,
                opt_run_property,
                supported_properties[opt_run_property],
                axes[1],
                start_indices,
                colors[j],
                legends[j],
                plot_type='hist',
            )
        else:
            axes = stats_lowlevel(
                result,
                opt_run_property,
                supported_properties[opt_run_property],
                axes,
                start_indices,
                colors[j],
                legends[j],
                plot_type,
            )

    if sum((legend is not None for legend in legends)) > 0:
        if plot_type == 'both':
            for ax in axes:
                ax.legend()
        else:
            axes.legend()

    return axes


def stats_lowlevel(
    result: Result,
    property_name: str,
    axis_label: str,
    ax: matplotlib.axes.Axes,
    start_indices: Optional[Union[int, Iterable[int]]] = None,
    color: Union[str, List[float], List[List[float]]] = 'C0',
    legend: Optional[str] = None,
    plot_type: str = 'line',
):
    """
    Plot values of the optimization run property across different multistarts.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'
    property_name:
        name of the optimization result property which value should be plotted
    axis_label:
        Label for the y axis of the line plot or x axis of the histogram
    ax:
        Axes object to use
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    color:
        List of RGBA colors (length equal to the number of multistarts),
        or single color, defined by a string or RGBA list
        If not set, then for the line plot clustering is done and
        colors are assigned automatically
    legend:
        Label describing the result
    plot_type:
        Specifies plot type. Possible values: 'line' and 'hist'

    Returns
    -------
    ax:
        The plot axes.
    """
    fvals = result.optimize_result.fval
    values = [res[property_name] for res in result.optimize_result.list]
    values, fvals = delete_nan_inf(fvals, values)

    if start_indices is not None:
        start_indices = process_start_indices(result, start_indices)
        values = values[start_indices]
        fvals = fvals[start_indices]

    n_starts = len(values)

    # assign colors
    colors = assign_colors(vals=fvals, colors=color, balance_alpha=False)

    # sort TODO: issue # 378
    sorted_indices = sorted(range(n_starts), key=lambda j: fvals[j])
    values = values[sorted_indices]

    if plot_type == 'line':
        # plot line
        ax.plot(range(n_starts), values, color=[0.7, 0.7, 0.7, 0.6])

        # plot points
        for i, v in enumerate(values):
            if i == 0:
                tmp_legend = legend
            else:
                tmp_legend = None
            ax.scatter(i, v, color=colors[i], marker='o', label=tmp_legend)
        ax.set_xlabel('Ordered optimizer run')
        ax.set_ylabel(axis_label)
    else:
        ax.hist(values, color=color, bins='auto', label=legend)
        ax.set_xlabel(axis_label)
        ax.set_ylabel('Number of multistarts')

    return ax



"""Validation intervals."""

import logging
from copy import deepcopy
from typing import Optional

from scipy.stats import chi2

from ..engine import Engine
from ..optimize import Optimizer, minimize
from ..problem import Problem
from ..result import Result

logger = logging.getLogger(__name__)


def validation_profile_significance(
    problem_full_data: Problem,
    result_training_data: Result,
    result_full_data: Optional[Result] = None,
    n_starts: Optional[int] = 1,
    optimizer: Optional[Optimizer] = None,
    engine: Optional[Engine] = None,
    lsq_objective: bool = False,
    return_significance: bool = True,
) -> float:
    r"""
    Compute significance of Validation Interval.

    It is a confidence region/interval for a new validation experiment.
    [#Kreutz]_ et al. (This method per default returns the significance =
    1-alpha!)

    The reasoning behind their approach is, that a validation data set
    is outside the validation interval, if fitting the full data set
    would lead to a fit $\theta_{new}$, that does not contain the old
    fit $\theta_{train}$ in their (Profile-Likelihood) based
    parameter-confidence intervals. (I.e. the old fit would be rejected by
    the fit of the full data.)

    This method returns the significance of the validation data set (where
    `result_full_data` is the objective function for fitting both data sets).
    I.e. the largest alpha, such that there is a validation region/interval
    such that the validation data set lies outside this Validation
    Interval with probability alpha. (If one is interested in the opposite,
    set `return_significance=False`.)

    Parameters
    ----------
    problem_full_data:
        pypesto.problem, such that the objective is the
        negative-log-likelihood of the training and validation data set.

    result_training_data:
        result object from the fitting of the training data set only.

    result_full_data
        pypesto.result object that contains the result of fitting
        training and validation data combined.

    n_starts
        number of starts for fitting the full data set
        (if result_full_data is not provided).

    optimizer:
        optimizer used for refitting the data (if result_full_data is not
        provided).

    engine
        engine for refitting (if result_full_data is not provided).

    lsq_objective:
        indicates if the objective of problem_full_data corresponds to a nllh
        (False), or a chi^2 value (True).
    return_significance:
        indicates, if the function should return the significance (True) (i.e.
        the  probability, that the new data set lies outside the Confidence
        Interval for the validation experiment, as given by the method), or
        the largest alpha, such that the validation experiment still lies
        within the Confidence Interval (False). I.e. alpha = 1-significance.


    .. [#Kreutz] Kreutz, Clemens, Raue, Andreas and Timmer, Jens.
        “Likelihood based observability analysis and
        confidence intervals for predictions of dynamic models”.
        BMC Systems Biology 2012/12. doi:10.1186/1752-0509-6-120
    """
    if (result_full_data is not None) and (optimizer is not None):
        raise UserWarning(
            "optimizer will not be used, as a result object "
            "for the full data set is provided."
        )

    # if result for full data is not provided: minimize
    if result_full_data is None:

        x_0 = result_training_data.optimize_result.x

        # copy problem, in order to not change/overwrite x_guesses
        problem = deepcopy(problem_full_data)
        problem.set_x_guesses(x_0)

        result_full_data = minimize(
            problem=problem,
            optimizer=optimizer,
            n_starts=n_starts,
            engine=engine,
            progress_bar=False,
        )

    # Validation intervals compare the nllh value on the full data set
    # of the parameter fits from the training and the full data set.

    nllh_new = result_full_data.optimize_result.fval[0]
    nllh_old = problem_full_data.objective(
        problem_full_data.get_reduced_vector(
            result_training_data.optimize_result.x[0]
        )
    )

    if nllh_new > nllh_old:
        logger.warning(
            "Fitting of the full data set provided a worse fit "
            "than the fit only considering the training data. "
            "Consider rerunning the not handing over "
            "result_full_data or running the fitting from the "
            "best parameters found from the training data."
        )

    # compute the probability, that the validation data set is outside the CI
    # => survival function chi.sf
    if return_significance:
        if lsq_objective:
            return chi2.sf(nllh_old - nllh_new, 1)
        else:
            return chi2.sf(2 * (nllh_old - nllh_new), 1)
    # compute the probability, that the validation data set is inside the CI
    # => cumulative density function chi.cdf
    else:
        if lsq_objective:
            return chi2.cdf(nllh_old - nllh_new, 1)
        else:
            return chi2.cdf(2 * (nllh_old - nllh_new), 1)



"""Miscellaneous methods."""
import logging
from typing import Dict, Iterable

import petab
import petab_select.ui
from petab.C import ESTIMATE, NOMINAL_VALUE
from petab_select import Model, parameter_string_to_value
from petab_select.constants import PETAB_PROBLEM

from ..objective import Objective
from ..petab import PetabImporter
from ..problem import Problem

logger = logging.getLogger(__name__)


def model_to_pypesto_problem(
    model: Model,
    objective: Objective = None,
    x_guesses: Iterable[Dict[str, float]] = None,
) -> Problem:
    """Create a pyPESTO problem from a PEtab Select model.

    Parameters
    ----------
    model:
        The model.
    objective:
        The pyPESTO objective.
    x_guesses:
        Startpoints to be used in the multi-start optimization. For example,
        this could be the maximum likelihood estimate from another model.
        Each dictionary has parameter IDs as keys, and parameter values as
        values.
        Values in `x_guess` for parameters that are not estimated will be
        ignored and replaced with their value from the PEtab Select model, if
        defined, else their nominal value in the PEtab parameters table.

    Returns
    -------
    Problem
        The pyPESTO select problem.
    """
    petab_problem = petab_select.ui.model_to_petab(model=model)[PETAB_PROBLEM]

    corrected_x_guesses = None
    if x_guesses is not None:
        corrected_x_guesses = correct_x_guesses(
            x_guesses=x_guesses,
            model=model,
            petab_problem=petab_problem,
        )

    importer = PetabImporter(petab_problem)
    if objective is None:
        amici_model = importer.create_model(
            non_estimated_parameters_as_constants=False,
        )
        objective = importer.create_objective(
            model=amici_model,
        )
    pypesto_problem = importer.create_problem(
        objective=objective,
        x_guesses=corrected_x_guesses,
    )
    return pypesto_problem


def correct_x_guesses(
    x_guesses: Iterable[Dict[str, float]],
    model: Model,
    petab_problem: petab.Problem = None,
):
    """Fix startpoint guesses passed between models of different sizes.

    Any parameter values in `x_guess` for parameters that are not estimated
    should be corrected by replacing them with
    - their corresponding values in `row` if possible, else
    - their corresponding nominal values in the `petab_problem.parameter_df`.
    """
    # TODO reconsider whether correcting is a good idea (`x_guess` is no longer
    # the latest MLE then). Similar todo exists in
    # `ModelSelectorMethod.new_model_problem`.
    # TODO move to PEtab Select?
    # TODO may be issues, e.g. differences in bounds of parameters between
    #      different model PEtab problems is not accounted for, or if there are
    #      different parameters in the old/new PEtab problem.

    if petab_problem is None:
        petab_problem = petab_select.ui.model_to_petab(model=model)[
            PETAB_PROBLEM
        ]

    corrected_x_guesses = None
    if x_guesses is not None:
        corrected_x_guesses = []
        for x_guess in x_guesses:
            corrected_x_guess = []
            for parameter_id in petab_problem.parameter_df.index:
                # Use the `x_guess` value, if the parameter is to be estimated.
                if (
                    petab_problem.parameter_df[ESTIMATE].loc[parameter_id] == 1
                    and parameter_id in x_guess
                ):
                    corrected_value = x_guess[parameter_id]
                # Else use the PEtab Select model parameter value, if defined.
                elif parameter_id in model.parameters:
                    corrected_value = parameter_string_to_value(
                        model.parameters[parameter_id]
                    )
                # Default to the PEtab parameter table nominal value
                else:
                    corrected_value = petab_problem.parameter_df.loc[
                        parameter_id, NOMINAL_VALUE
                    ]
                corrected_x_guess.append(corrected_value)
            corrected_x_guesses.append(corrected_x_guess)
    return corrected_x_guesses



"""
Constants
=========
Package-wide consistent constant definitions.
"""

from enum import Enum
from typing import Callable, Literal, Tuple, Union

###############################################################################
# ENSEMBLE

PREDICTOR = 'predictor'
PREDICTION_ID = 'prediction_id'
PREDICTION_RESULTS = 'prediction_results'
PREDICTION_ARRAYS = 'prediction_arrays'
PREDICTION_SUMMARY = 'prediction_summary'

HISTORY = 'history'
OPTIMIZE = 'optimize'
SAMPLE = 'sample'

MEAN = 'mean'
MEDIAN = 'median'
STANDARD_DEVIATION = 'std'
PERCENTILE = 'percentile'
SUMMARY = 'summary'
WEIGHTED_SIGMA = 'weighted_sigma'

X_NAMES = 'x_names'
NX = 'n_x'
X_VECTOR = 'x_vectors'
NVECTORS = 'n_vectors'
VECTOR_TAGS = 'vector_tags'
ENSEMBLE_TYPE = 'ensemble_type'
PREDICTIONS = 'predictions'

LOWER_BOUND = 'lower_bound'
UPPER_BOUND = 'upper_bound'
PREEQUILIBRATION_CONDITION_ID = 'preequilibrationConditionId'
SIMULATION_CONDITION_ID = 'simulationConditionId'

COLOR_HIT_BOTH_BOUNDS = [0.6, 0.0, 0.0, 0.9]
COLOR_HIT_ONE_BOUND = [0.95, 0.6, 0.0, 0.9]
COLOR_HIT_NO_BOUNDS = [0.0, 0.8, 0.0, 0.9]


class EnsembleType(Enum):
    """Specifies different ensemble types."""

    ensemble = 1
    sample = 2
    unprocessed_chain = 3


###############################################################################
# OBJECTIVE

MODE_FUN = 'mode_fun'  # mode for function values
MODE_RES = 'mode_res'  # mode for residuals
ModeType = Literal['mode_fun', 'mode_res']  # type for `mode` argument
FVAL = 'fval'  # function value
FVAL0 = 'fval0'  # function value at start
GRAD = 'grad'  # gradient
HESS = 'hess'  # Hessian
HESSP = 'hessp'  # Hessian vector product
RES = 'res'  # residual
SRES = 'sres'  # residual sensitivities
RDATAS = 'rdatas'  # returned simulated data sets

TIME = 'time'  # time
N_FVAL = 'n_fval'  # number of function evaluations
N_GRAD = 'n_grad'  # number of gradient evaluations
N_HESS = 'n_hess'  # number of Hessian evaluations
N_RES = 'n_res'  # number of residual evaluations
N_SRES = 'n_sres'  # number of residual sensitivity evaluations
START_TIME = 'start_time'  # start time
X = 'x'
X0 = 'x0'
ID = 'id'


###############################################################################
# HIERARCHICAL

INNER_PARAMETERS = 'inner_parameters'
INNER_RDATAS = 'inner_rdatas'
PARAMETER_TYPE = 'parameterType'


class InnerParameterType(str, Enum):
    """Specifies different inner parameter types."""

    OFFSET = 'offset'
    SCALING = 'scaling'
    SIGMA = 'sigma'


DUMMY_INNER_VALUE = {
    InnerParameterType.OFFSET: 0.0,
    InnerParameterType.SCALING: 1.0,
    InnerParameterType.SIGMA: 1.0,
}

INNER_PARAMETER_BOUNDS = {
    InnerParameterType.OFFSET: {
        LOWER_BOUND: -float('inf'),
        UPPER_BOUND: float('inf'),
    },
    InnerParameterType.SCALING: {
        LOWER_BOUND: -float('inf'),
        UPPER_BOUND: float('inf'),
    },
    InnerParameterType.SIGMA: {
        LOWER_BOUND: 0,
        UPPER_BOUND: float('inf'),
    },
}


###############################################################################
# HISTORY

HISTORY = "history"
TRACE = "trace"
N_ITERATIONS = "n_iterations"
MESSAGES = "messages"
MESSAGE = "message"
EXITFLAG = "exitflag"
TRACE_SAVE_ITER = "trace_save_iter"

SUFFIXES_CSV = ["csv"]
SUFFIXES_HDF5 = ["hdf5", "h5"]
SUFFIXES = SUFFIXES_CSV + SUFFIXES_HDF5


###############################################################################
# PRIOR

LIN = 'lin'  # linear
LOG = 'log'  # logarithmic to basis e
LOG10 = 'log10'  # logarithmic to basis 10

UNIFORM = 'uniform'
PARAMETER_SCALE_UNIFORM = 'parameterScaleUniform'
NORMAL = 'normal'
PARAMETER_SCALE_NORMAL = 'parameterScaleNormal'
LAPLACE = 'laplace'
PARAMETER_SCALE_LAPLACE = 'parameterScaleLaplace'
LOG_UNIFORM = 'logUniform'
LOG_NORMAL = 'logNormal'
LOG_LAPLACE = 'logLaplace'


###############################################################################
# PREDICT

OUTPUT_IDS = 'output_ids'  # data member in PredictionConditionResult
PARAMETER_IDS = 'x_names'  # data member in PredictionConditionResult
TIMEPOINTS = 'timepoints'  # data member in PredictionConditionResult
OUTPUT = 'output'  # field in the return dict of AmiciPredictor
OUTPUT_SENSI = 'output_sensi'  # field in the return dict of AmiciPredictor
OUTPUT_WEIGHT = 'output_weight'  # field in the return dict of AmiciPredictor
OUTPUT_SIGMAY = 'output_sigmay'  # field in the return dict of AmiciPredictor

# separator in the conditions_ids between preequilibration and simulation
# condition
CONDITION_SEP = '::'

AMICI_T = 't'  # return field in amici simulation result
AMICI_X = 'x'  # return field in amici simulation result
AMICI_SX = 'sx'  # return field in amici simulation result
AMICI_Y = 'y'  # return field in amici simulation result
AMICI_SY = 'sy'  # return field in amici simulation result
AMICI_LLH = 'llh'  # return field in amici simulation result
AMICI_STATUS = 'status'  # return field in amici simulation result
AMICI_SIGMAY = 'sigmay'  # return field in amici simulation result

CONDITION = 'condition'
CONDITION_IDS = 'condition_ids'

CSV = 'csv'  # return file format
H5 = 'h5'  # return file format


###############################################################################
# SELECT

TYPE_POSTPROCESSOR = Callable[["ModelProblem"], None]  # noqa: F821


###############################################################################
# VISUALIZE

LEN_RGB = 3  # number of elements in an RGB color
LEN_RGBA = 4  # number of elements in an RGBA color
RGB = Tuple[(float,) * LEN_RGB]  # typing of an RGB color
RGBA = Tuple[(float,) * LEN_RGBA]  # typing of an RGBA color
RGB_RGBA = Union[RGB, RGBA]  # typing of an RGB or RGBA color
RGBA_MIN = 0  # min value for an RGBA element
RGBA_MAX = 1  # max value for an RGBA element
RGBA_ALPHA = 3  # zero-indexed fourth element in RGBA
RGBA_WHITE = (RGBA_MAX, RGBA_MAX, RGBA_MAX, RGBA_MAX)  # white as an RGBA color
RGBA_BLACK = (RGBA_MIN, RGBA_MIN, RGBA_MIN, RGBA_MAX)  # black as an RGBA color

# optimizer history
TRACE_X_TIME = 'time'
TRACE_X_STEPS = 'steps'
# supported values to plot on x-axis
TRACE_X = (TRACE_X_TIME, TRACE_X_STEPS)

TRACE_Y_FVAL = 'fval'
TRACE_Y_GRADNORM = 'gradnorm'
# supported values to plot on y-axis
TRACE_Y = (TRACE_Y_FVAL, TRACE_Y_GRADNORM)

# parameter indices
FREE_ONLY = 'free_only'  # only estimated parameters
ALL = 'all'  # all parameters, also for start indices

# start indices
ALL_CLUSTERED = 'all_clustered'  # best + all that are in a cluster of size > 1
FIRST_CLUSTER = 'first_cluster'  # all starts that belong to the first cluster


###############################################################################
# ENVIRONMENT VARIABLES

PYPESTO_MAX_N_STARTS: str = "PYPESTO_MAX_N_STARTS"
PYPESTO_MAX_N_SAMPLES: str = "PYPESTO_MAX_N_SAMPLES"



import warnings
from typing import List

import numpy as np

from ..C import DUMMY_INNER_VALUE, InnerParameterType


def get_finite_quotient(
    numerator: float,
    denominator: float,
    inner_parameter_type: InnerParameterType,
):
    """Get a finite value for the inner parameter.

    Parameters
    ----------
    numerator:
        The numerator of a quotient.
    denominator:
        The denominator of a quotient.
    useless:
        The value to return if the quotient is not finite.

    Returns
    -------
    `num / den` if it's finite, else a dummy value.
    """
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            quotient = float(numerator / denominator)
        if not np.isfinite(quotient):
            raise ValueError
        return quotient
    except Exception:
        return DUMMY_INNER_VALUE[inner_parameter_type]


def compute_optimal_scaling(
    data: List[np.ndarray],
    sim: List[np.ndarray],
    sigma: List[np.ndarray],
    mask: List[np.ndarray],
) -> float:
    """
    Compute optimal scaling.

    Compute optimal scaling parameter for the given measurements and model
    outputs. See https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.1
    for the derivation.
    """
    # SI Section 3.1: page 9, bottom (s = ...) with offset b=0 (as there is
    # no offset, or it already has been subtracted from the measurements)
    # numerator and denominator to compute the optimal scaling
    # num: \sum_i \frac{ \bar{y}_i * \tilde{h}_i } { \sigma_i^2 }
    # den:  \sum_i \frac{ \tilde{h}_i^2 }{ \sigma_i^2 }

    # numerator, denominator
    num, den = 0.0, 0.0

    # iterate over conditions
    for sim_i, data_i, sigma_i, mask_i in zip(sim, data, sigma, mask):
        # extract relevant values
        sim_x = sim_i[mask_i]  # \tilde{h}_i
        data_x = data_i[mask_i]  # \bar{y}_i
        sigma_x = sigma_i[mask_i]  # \sigma_i
        # update statistics
        num += np.nansum(sim_x * data_x / sigma_x**2)
        den += np.nansum(sim_x**2 / sigma_x**2)

    return get_finite_quotient(
        numerator=num,
        denominator=den,
        inner_parameter_type=InnerParameterType.SCALING,
    )


def apply_scaling(
    scaling_value: float, sim: List[np.ndarray], mask: List[np.ndarray]
):
    """Apply scaling to simulations (in-place).

    Parameters
    ----------
    scaling_value:
        The optimal offset for the masked simulations.
    sim:
        All full (unmasked) simulations.
    mask:
        The masks that indicate the simulation subset that corresponds to the
        `scaling_value`.
    """
    for i in range(len(sim)):
        sim[i][mask[i]] *= scaling_value


def compute_optimal_offset(
    data: List[np.ndarray],
    sim: List[np.ndarray],
    sigma: List[np.ndarray],
    mask: List[np.ndarray],
) -> float:
    """Compute optimal offset.

    Compute optimal offset for the given measurements and model outputs. See
    https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.1 for the
    derivation. This function handles offsets that occur without any coupled
    scaling parameter.
    """
    # SI Section 3.1: page 9, bottom (b = ...) with scaling s=1
    # numerator and denominator to compute the optimal offset
    # num: \sum_i \frac{ \bar{y}_i - \tilde{h}_i } { \sigma_i^2 }
    # den:  \sum_i \frac{1}{ \sigma_i^2 }
    num, den = 0.0, 0.0

    # iterate over conditions
    for sim_i, data_i, sigma_i, mask_i in zip(sim, data, sigma, mask):
        # extract relevant values
        sim_x = sim_i[mask_i]  # \tilde{h}_i
        data_x = data_i[mask_i]  # \bar{y}_i
        sigma_x = sigma_i[mask_i]  # \sigma_i
        # update statistics
        num += np.nansum((data_x - sim_x) / sigma_x**2)
        den += np.nansum(1 / sigma_x**2)

    return get_finite_quotient(
        numerator=num,
        denominator=den,
        inner_parameter_type=InnerParameterType.OFFSET,
    )


def compute_optimal_offset_coupled(
    data: List[np.ndarray],
    sim: List[np.ndarray],
    sigma: List[np.ndarray],
    mask: List[np.ndarray],
) -> float:
    """Compute optimal offset.

    Compute optimal offset for an observable that has both an offset and
    scaling inner parameter.

    See https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.1 for the
    derivation.
    """
    # will be \sum_i \frac{ \tilde{h}_i }{ \sigma_i^2 }
    h = 0.0
    # will be \sum_i \frac{1}{ \sigma_i^2 }
    recnoise = 0.0
    # will be \sum_i \frac{ \bar{y}_i * \tilde{h}_i }{ \sigma_i^2 }
    yh = 0.0
    # will be \sum_i \frac{ \bar{y}_i }{ \sigma_i^2 }
    y = 0.0
    # will be \sum_i \frac{ \tilde{h}_i^2}{ \sigma_i^2 }
    h2 = 0.0

    # iterate over conditions
    for sim_i, data_i, sigma_i, mask_i in zip(sim, data, sigma, mask):
        if mask_i.max(initial=False) is False:
            continue
        # extract relevant values
        sim_x = sim_i[mask_i]  # \tilde{h}_i
        data_x = data_i[mask_i]  # \bar{y}_i
        sigma_x = sigma_i[mask_i]  # \sigma_i
        # update statistics
        s2 = sigma_x**2  # \sigma_i^2
        h += np.nansum(sim_x / s2)
        recnoise += np.nansum(1 / s2)
        yh += np.nansum((sim_x * data_x) / s2)
        y += np.nansum(data_x / s2)
        h2 += np.nansum((sim_x**2) / s2)

    # numerator and denominator in equation 11, each multiplied by
    #  recnoise = \sum_i \frac{1}{ \sigma_i^2 }
    num = y - (yh * h) / h2
    den = recnoise - (h**2) / h2

    # If simulation is essentially constant, then offset and scaling
    # have the same effect. In this case, we set the offset to a dummy value,
    # and estimate scaling. This is because offset is computed first, by the
    # calculator.
    # NB: this will cause issues if e.g. a simulation with dynamics has values
    #     on the order of the zero check here. An attempt to avoid this
    #     is done here by checking that the numerator is not close to zero.
    if np.isclose(den, 0, atol=1e-14) and not np.isclose(num, 0, atol=1e-4):
        # `get_finite_quotient` will now return the dummy value
        den = 0

    return get_finite_quotient(
        numerator=num,
        denominator=den,
        inner_parameter_type=InnerParameterType.OFFSET,
    )


def apply_offset(
    offset_value: float, data: List[np.ndarray], mask: List[np.ndarray]
):
    """Apply offset to data (in-place).

    Acts on data instead of simulation because of:
    `data = scaling * simulation + offset`. The first step of transforming
    data and simulation so they can be compared is to either multiply
    `simulation` by `scaling`, or subtract `offset` from `data`.
    As we only currently have `compute_optimal_offset_coupled`, not
    `compute_optimal_scaling_coupled`, we need to compute optimal offset first,
    so the first step must act on data.

    NB: as this applies to data in-place, it's important to supply a copy of
    data s.t. the original data is not affected.

    Parameters
    ----------
    offset_value:
        The optimal offset for the masked simulations.
    sim:
        All full (unmasked) data.
    mask:
        The masks that indicate the data subset that corresponds to the
        `offset_value`.
    """
    for i in range(len(data)):
        data[i][mask[i]] -= offset_value


def compute_optimal_sigma(
    data: List[np.ndarray], sim: List[np.ndarray], mask: List[np.ndarray]
) -> float:
    """Compute optimal sigma.

    Compute optimal sigmas for the given measurements and model outputs. See
    https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.2 for the
    derivation.
    """
    # numerator, denominator
    num, den = 0.0, 0.0

    # iterate over conditions
    for sim_i, data_i, mask_i in zip(sim, data, mask):
        # extract relevant values
        sim_x = sim_i[mask_i]
        data_x = data_i[mask_i]
        # update statistics
        num += np.nansum((data_x - sim_x) ** 2)
        den += sim_x.size

    if num == 0:
        raise AssertionError()
    # compute optimal value
    return np.sqrt(num / den)


def apply_sigma(
    sigma_value: float, sigma: List[np.ndarray], mask: List[np.ndarray]
):
    """Apply optimal sigma to pre-existing sigma arrays (in-place).

    Parameters
    ----------
    sigma_value:
        The optimal sigma value.
    sigma:
        All full (unmasked) sigmas.
    mask:
        The masks that indicate the sigma subset that corresponds to the
        `sigma_value`.
    """
    for i in range(len(sigma)):
        sigma[i][mask[i]] = sigma_value


def compute_nllh(
    data: List[np.ndarray], sim: List[np.ndarray], sigma: List[np.ndarray]
) -> float:
    """Compute negative log-likelihood.

    Compute negative log-likelihood of the data, given the model outputs and
    sigmas.
    """
    return sum(
        0.5 * np.nansum(np.log(2 * np.pi * sigma_i**2))
        + 0.5 * np.nansum((data_i - sim_i) ** 2 / sigma_i**2)
        for data_i, sim_i, sigma_i in zip(data, sim, sigma)
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Sequence, Tuple, Union

import numpy as np

from ...C import (
    FVAL,
    GRAD,
    HESS,
    MODE_FUN,
    MODE_RES,
    RDATAS,
    RES,
    SRES,
    ModeType,
)
from .amici_util import (
    add_sim_grad_to_opt_grad,
    add_sim_hess_to_opt_hess,
    filter_return_dict,
    get_error_output,
    init_return_values,
    log_simulation,
    sim_sres_to_opt_sres,
)

if TYPE_CHECKING:
    try:
        import amici
        from amici.parameter_mapping import ParameterMapping
    except ImportError:
        ParameterMapping = None

AmiciModel = Union['amici.Model', 'amici.ModelPtr']
AmiciSolver = Union['amici.Solver', 'amici.SolverPtr']


class AmiciCalculator:
    """Class to perform the AMICI call and obtain objective function values."""

    def __init__(self):
        self._known_least_squares_safe = False

    def initialize(self):
        """Initialize the calculator. Default: Do nothing."""

    def __call__(
        self,
        x_dct: Dict,
        sensi_orders: Tuple[int],
        mode: ModeType,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: List[amici.ExpData],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
    ):
        """Perform the actual AMICI call.

        Called within the :func:`AmiciObjective.__call__` method.

        Parameters
        ----------
        x_dct:
            Parameters for which to compute function value and derivatives.
        sensi_orders:
            Tuple of requested sensitivity orders.
        mode:
            Call mode (function value or residual based).
        amici_model:
            The AMICI model.
        amici_solver:
            The AMICI solver.
        edatas:
            The experimental data.
        n_threads:
            Number of threads for AMICI call.
        x_ids:
            Ids of optimization parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        fim_for_hess:
            Whether to use the FIM (if available) instead of the Hessian (if
            requested).
        """
        import amici.parameter_mapping

        # set order in solver
        sensi_order = 0
        if sensi_orders:
            sensi_order = max(sensi_orders)

        if sensi_order == 2 and fim_for_hess:
            # we use the FIM
            amici_solver.setSensitivityOrder(sensi_order - 1)
        else:
            amici_solver.setSensitivityOrder(sensi_order)

        # fill in parameters
        amici.parameter_mapping.fill_in_parameters(
            edatas=edatas,
            problem_parameters=x_dct,
            scaled_parameters=True,
            parameter_mapping=parameter_mapping,
            amici_model=amici_model,
        )

        # run amici simulation
        rdatas = amici.runAmiciSimulations(
            amici_model,
            amici_solver,
            edatas,
            num_threads=min(n_threads, len(edatas)),
        )
        if (
            not self._known_least_squares_safe
            and mode == MODE_RES
            and 1 in sensi_orders
        ):
            if not amici_model.getAddSigmaResiduals() and any(
                (
                    (r['ssigmay'] is not None and np.any(r['ssigmay']))
                    or (r['ssigmaz'] is not None and np.any(r['ssigmaz']))
                )
                for r in rdatas
            ):
                raise RuntimeError(
                    'Cannot use least squares solver with'
                    'parameter dependent sigma! Support can be '
                    'enabled via '
                    'amici_model.setAddSigmaResiduals().'
                )
            self._known_least_squares_safe = True  # don't check this again

        return calculate_function_values(
            rdatas=rdatas,
            sensi_orders=sensi_orders,
            mode=mode,
            amici_model=amici_model,
            amici_solver=amici_solver,
            edatas=edatas,
            x_ids=x_ids,
            parameter_mapping=parameter_mapping,
            fim_for_hess=fim_for_hess,
        )


def calculate_function_values(
    rdatas,
    sensi_orders: Tuple[int, ...],
    mode: ModeType,
    amici_model: AmiciModel,
    amici_solver: AmiciSolver,
    edatas: List[amici.ExpData],
    x_ids: Sequence[str],
    parameter_mapping: ParameterMapping,
    fim_for_hess: bool,
):
    """Calculate the function values from rdatas and return as dict."""
    import amici

    # full optimization problem dimension (including fixed parameters)
    dim = len(x_ids)

    # check if the simulation failed
    if any(rdata['status'] < 0.0 for rdata in rdatas):
        return get_error_output(
            amici_model, edatas, rdatas, sensi_orders, mode, dim
        )

    nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
        sensi_orders, mode, dim
    )

    par_sim_ids = list(amici_model.getParameterIds())
    sensi_method = amici_solver.getSensitivityMethod()

    # iterate over return data
    for data_ix, rdata in enumerate(rdatas):
        log_simulation(data_ix, rdata)

        condition_map_sim_var = parameter_mapping[data_ix].map_sim_var

        # add objective value
        nllh -= rdata['llh']

        if mode == MODE_FUN:
            if not np.isfinite(nllh):
                return get_error_output(
                    amici_model, edatas, rdatas, sensi_orders, mode, dim
                )

            if 1 in sensi_orders:
                # add gradient
                add_sim_grad_to_opt_grad(
                    x_ids,
                    par_sim_ids,
                    condition_map_sim_var,
                    rdata['sllh'],
                    snllh,
                    coefficient=-1.0,
                )

                if not np.isfinite(snllh).all():
                    return get_error_output(
                        amici_model, edatas, rdatas, sensi_orders, mode, dim
                    )

                # Hessian
            if 2 in sensi_orders:
                if (
                    sensi_method != amici.SensitivityMethod_forward
                    or not fim_for_hess
                ):
                    raise ValueError("AMICI cannot compute Hessians yet.")
                    # add FIM for Hessian
                add_sim_hess_to_opt_hess(
                    x_ids,
                    par_sim_ids,
                    condition_map_sim_var,
                    rdata['FIM'],
                    s2nllh,
                    coefficient=+1.0,
                )
                if not np.isfinite(s2nllh).all():
                    return get_error_output(
                        amici_model, edatas, rdatas, sensi_orders, mode, dim
                    )

        elif mode == MODE_RES:
            if 0 in sensi_orders:
                chi2 += rdata['chi2']
                res = (
                    np.hstack([res, rdata['res']])
                    if res.size
                    else rdata['res']
                )
            if 1 in sensi_orders:
                opt_sres = sim_sres_to_opt_sres(
                    x_ids,
                    par_sim_ids,
                    condition_map_sim_var,
                    rdata['sres'],
                    coefficient=1.0,
                )
                sres = np.vstack([sres, opt_sres]) if sres.size else opt_sres

    ret = {
        FVAL: nllh,
        GRAD: snllh,
        HESS: s2nllh,
        RES: res,
        SRES: sres,
        RDATAS: rdatas,
    }

    return filter_return_dict(ret)



"""Engines with multi-process parallelization."""
import logging
import os
from multiprocessing import Pool
from typing import List

import cloudpickle as pickle
from tqdm import tqdm

from .base import Engine
from .task import Task

logger = logging.getLogger(__name__)


def work(pickled_task):
    """Unpickle and execute task."""
    task = pickle.loads(pickled_task)
    return task.execute()


class MultiProcessEngine(Engine):
    """
    Parallelize the task execution using multiprocessing.

    Parameters
    ----------
    n_procs:
        The maximum number of processes to use in parallel.
        Defaults to the number of CPUs available on the system according to
        `os.cpu_count()`.
        The effectively used number of processes will be the minimum of
        `n_procs` and the number of tasks submitted.
    """

    def __init__(self, n_procs: int = None):
        super().__init__()

        if n_procs is None:
            n_procs = os.cpu_count()
            logger.warning(
                f"Engine set up to use up to {n_procs} processes in total. "
                f"The number was automatically determined and might not be "
                f"appropriate on some systems."
            )
        self.n_procs: int = n_procs

    def execute(self, tasks: List[Task], progress_bar: bool = True):
        """Pickle tasks and distribute work over parallel processes.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        n_tasks = len(tasks)

        pickled_tasks = [pickle.dumps(task) for task in tasks]

        n_procs = min(self.n_procs, n_tasks)
        logger.info(
            f"Performing parallel task execution on {n_procs} " f"processes."
        )

        with Pool(processes=n_procs) as pool:
            results = pool.map(
                work, tqdm(pickled_tasks, disable=not progress_bar)
            )

        return results



from __future__ import annotations

import copy
from typing import Dict, List, Optional, Sequence, Tuple

import amici
import numpy as np
from amici.parameter_mapping import ParameterMapping

from ..C import (
    AMICI_SIGMAY,
    AMICI_Y,
    GRAD,
    INNER_PARAMETERS,
    INNER_RDATAS,
    RDATAS,
    ModeType,
)
from ..objective.amici.amici_calculator import (
    AmiciCalculator,
    AmiciModel,
    AmiciSolver,
)
from .problem import AmiciInnerProblem
from .solver import AnalyticalInnerSolver, InnerSolver


class HierarchicalAmiciCalculator(AmiciCalculator):
    """
    A calculator that is passed as `calculator` to the pypesto.AmiciObjective.

    While this class cannot be used directly, it has two subclasses
    which allow to use forward or adjoint sensitivity analysis to
    solve a `pypesto.HierarchicalProblem` efficiently in an inner loop,
    while the outer optimization is only concerned with variables not
    specified as `pypesto.HierarchicalParameter`s.
    """

    def __init__(
        self,
        inner_problem: AmiciInnerProblem,
        inner_solver: Optional[InnerSolver] = None,
    ):
        """Initialize the calculator from the given problem.

        Arguments
        ---------
        inner_problem:
            The inner problem of a hierarchical optimization problem.
        inner_solver:
            A solver to solve ``inner_problem``.
            Defaults to ``pypesto.hierarchical.solver.AnalyticalInnerSolver``.
        """
        super().__init__()

        self.inner_problem = inner_problem

        if inner_solver is None:
            inner_solver = AnalyticalInnerSolver()
        self.inner_solver = inner_solver

    def initialize(self):
        """Initialize."""
        super().initialize()
        self.inner_solver.initialize()

    def __call__(
        self,
        x_dct: Dict,
        sensi_orders: Tuple[int],
        mode: ModeType,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: List[amici.ExpData],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
    ):
        """Perform the actual AMICI call, with hierarchical optimization.

        See documentation for
        `pypesto.objective.amici.amici_calculator.AmiciCalculator.__call__()`.

        The return object also includes the simulation results that were
        generated to solve the inner problem, as well as the parameters that
        solver the inner problem.
        """
        if not self.inner_problem.check_edatas(edatas=edatas):
            raise ValueError(
                'The experimental data provided to this call differs from '
                'the experimental data used to setup the hierarchical '
                'optimizer.'
            )

        # compute optimal inner parameters
        x_dct = copy.deepcopy(x_dct)
        x_dct.update(self.inner_problem.get_dummy_values(scaled=True))
        inner_result = super().__call__(
            x_dct=x_dct,
            sensi_orders=(0,),
            mode=mode,
            amici_model=amici_model,
            amici_solver=amici_solver,
            edatas=edatas,
            n_threads=n_threads,
            x_ids=x_ids,
            parameter_mapping=parameter_mapping,
            fim_for_hess=fim_for_hess,
        )
        inner_rdatas = inner_result[RDATAS]

        # if any amici simulation failed, it's unlikely we can compute
        # meaningful inner parameters, so we better just fail early.
        if any(rdata.status != amici.AMICI_SUCCESS for rdata in inner_rdatas):
            # if the gradient was requested, we need to provide some value
            # for it
            if 1 in sensi_orders:
                inner_result[GRAD] = np.full(
                    shape=len(x_ids), fill_value=np.nan
                )
            return inner_result

        inner_parameters = self.inner_solver.solve(
            problem=self.inner_problem,
            sim=[rdata[AMICI_Y] for rdata in inner_rdatas],
            sigma=[rdata[AMICI_SIGMAY] for rdata in inner_rdatas],
            scaled=True,
        )

        # fill in optimal values
        # directly writing to parameter mapping ensures that plists do not
        # include hierarchically computed parameters
        x_dct = copy.deepcopy(x_dct)
        x_dct.update(inner_parameters)

        # TODO use plist to compute only required derivatives, in
        #  `super.__call__`, `amici.parameter_mapping.fill_in_parameters`
        # TODO speed gain: if no gradient is required, then simulation can be
        #  skipped here, and rdatas can be updated in place
        #  (y, sigma, res, llh).
        result = super().__call__(
            x_dct=x_dct,
            sensi_orders=sensi_orders,
            mode=mode,
            amici_model=amici_model,
            amici_solver=amici_solver,
            edatas=edatas,
            n_threads=n_threads,
            x_ids=x_ids,
            parameter_mapping=parameter_mapping,
            fim_for_hess=fim_for_hess,
        )
        result[INNER_PARAMETERS] = inner_parameters
        result[INNER_RDATAS] = inner_rdatas

        return result



"""Utility function for profile module."""
from typing import Any, Dict, Iterable, Tuple

import numpy as np
import scipy.stats

from ..C import GRAD
from ..problem import Problem
from ..result import ProfileResult, ProfilerResult, Result


def chi2_quantile_to_ratio(alpha: float = 0.95, df: int = 1):
    """
    Compute profile likelihood threshold.

    Transform lower tail probability `alpha` for a chi2 distribution with `df`
    degrees of freedom to a profile likelihood ratio threshold.

    Parameters
    ----------
    alpha:
        Lower tail probability, defaults to 95% interval.
    df:
        Degrees of freedom.

    Returns
    -------
    ratio:
        Corresponds to a likelihood ratio.
    """
    quantile = scipy.stats.chi2.ppf(alpha, df=df)
    ratio = np.exp(-quantile / 2)
    return ratio


def calculate_approximate_ci(
    xs: np.ndarray, ratios: np.ndarray, confidence_ratio: float
) -> Tuple[float, float]:
    """
    Calculate approximate confidence interval based on profile.

    Interval bounds are linerly interpolated.

    Parameters
    ----------
    xs:
        The ordered parameter values along the profile for the coordinate of
        interest.
    ratios:
        The likelihood ratios corresponding to the parameter values.
    confidence_ratio:
        Minimum confidence ratio to base the confidence interval upon, as
        obtained via `pypesto.profile.chi2_quantile_to_ratio`.

    Returns
    -------
    lb, ub:
        Bounds of the approximate confidence interval.
    """
    # extract indices where the ratio is larger than the minimum ratio
    (indices,) = np.where(ratios >= confidence_ratio)
    l_ind, u_ind = indices[0], indices[-1]

    # lower bound
    if l_ind == 0:
        lb = xs[l_ind]
    else:
        # linear interpolation with next smaller value
        ind = [l_ind - 1, l_ind]
        lb = np.interp(confidence_ratio, ratios[ind], xs[ind])

    # upper bound
    if u_ind == len(ratios) - 1:
        ub = xs[u_ind]
    else:
        # linear interpolation with next larger value
        ind = [u_ind + 1, u_ind]  # flipped as interp expects increasing xs
        ub = np.interp(confidence_ratio, ratios[ind], xs[ind])

    return lb, ub


def initialize_profile(
    problem: Problem,
    result: Result,
    result_index: int,
    profile_index: Iterable[int],
    profile_list: int,
) -> float:
    """
    Initialize profiling based on a previous optimization.

    Parameters
    ----------
    problem:
        The problem to be solved.
    result:
        A result object to initialize profiling and to append the profiling
        results to. For example, one might append more profiling runs to a
        previous profile, in order to merge these.
        The existence of an optimization result is obligatory.
    result_index:
        index from which optimization result profiling should be started
    profile_index:
        array with parameter indices, whether a profile should
        be computed (1) or not (0)
        Default is all profiles should be computed
    profile_list:
        integer which specifies whether a call to the profiler should create
        a new list of profiles (default) or should be added to a specific
        profile list

    Returns
    -------
    global_opt:
        log-posterior at global optimum.
    """
    # Check whether an optimization result is existing
    if result.optimize_result is None:
        raise ValueError(
            "Optimization has to be carried out before profiling can be done."
        )

    tmp_optimize_result = result.optimize_result.as_list()

    # Check if new profile_list is to be created
    if profile_list is None:
        result.profile_result.append_empty_profile_list()

    # get the log-posterior of the global optimum
    global_opt = tmp_optimize_result[0]["fval"]

    # fill the list with optimization results where necessary
    fill_profile_list(
        profile_result=result.profile_result,
        optimizer_result=tmp_optimize_result[result_index],
        profile_index=profile_index,
        profile_list=profile_list,
        problem_dimension=problem.dim_full,
        global_opt=global_opt,
    )

    # return the log-posterior of the global optimum (needed in order to
    # compute the log-posterior-ratio)
    return global_opt


def fill_profile_list(
    profile_result: ProfileResult,
    optimizer_result: Dict[str, Any],
    profile_index: Iterable[int],
    profile_list: int,
    problem_dimension: int,
    global_opt: float,
) -> None:
    """Fill a ProfileResult.

    Helper function for `initialize_profile`.

    Parameters
    ----------
    profile_result:
        A list of profiler result objects.
    optimizer_result:
        A local optimization result.
    profile_index:
        array with parameter indices, whether a profile should
        be computed (1) or not (0).
        Default is all profiles should be computed.
    profile_list:
        integer which specifies whether a call to the profiler should
        create a new list of profiles (default) or should be added to a
        specific profile list.
    problem_dimension:
        number of parameters in the unreduced problem.
    global_opt:
        log-posterior at global optimum.
    """
    if optimizer_result[GRAD] is not None:
        gradnorm = np.linalg.norm(optimizer_result[GRAD])
    else:
        gradnorm = np.nan

    # create blank profile
    new_profile = ProfilerResult(
        x_path=optimizer_result["x"],
        fval_path=np.array([optimizer_result["fval"]]),
        ratio_path=np.array([np.exp(global_opt - optimizer_result["fval"])]),
        gradnorm_path=gradnorm,
        exitflag_path=optimizer_result["exitflag"],
        time_path=np.array([0.0]),
        time_total=0.0,
        n_fval=0,
        n_grad=0,
        n_hess=0,
        message=None,
    )

    if profile_list is None:
        # All profiles have to be created from scratch
        for i_parameter in range(0, problem_dimension):
            if i_parameter in profile_index:
                # Should we create a profile for this index?
                profile_result.append_profiler_result(new_profile)
            else:
                # if no profile should be computed for this parameter
                profile_result.append_profiler_result(None)

    else:
        for i_parameter in range(0, problem_dimension):
            # We append to an existing list
            if i_parameter in profile_index:
                # Do we have to create a new profile?
                create_new = (
                    profile_result.list[profile_list][i_parameter] is None
                )
                if create_new:
                    profile_result.set_profiler_result(
                        new_profile, i_parameter
                    )
